try:
  import cPickle as pickle
except ImportError:
  import pickle

from collections import defaultdict
from contextlib import contextmanager
import os

from twitter.common.collections import OrderedSet
from twitter.common.dirutil import safe_rmtree, safe_mkdir, safe_open

from twitter.pants.base import manual
from twitter.pants.base.build_cache import BuildCache, NO_SOURCES, TARGET_SOURCES
from twitter.pants.targets.external_dependency import ExternalDependency


@manual.builddict()
class TaskError(Exception):
  """
    Raised to indicate a task has failed.
  
    :param int exit_code: an optional exit code (1, by default)
  """
  def __init__(self, *args, **kwargs):
    self._exit_code = kwargs.pop('exit_code', 1)
    super(TaskError, self).__init__(*args, **kwargs)

  @property
  def exit_code(self):
    return self._exit_code

class TargetError(TaskError):
  """
    Raised to indicate a task has failed for a subset of targets.
    
    :param list targets: A list of pants targets that failed
  """
  def __init__(self, targets, *args, **kwargs):
    TaskError.__init__(self, *args, **kwargs)
    self.targets = targets


@manual.builddict()
class Task(object):
  """Base class for all Tasks."""

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    """
      Subclasses can add flags to the pants command line using the given option group.  Flag names
      should be created with mkflag([name]) to ensure flags are properly namespaced amongst other
      tasks.
    """

  EXTRA_DATA = 'extra.data'

  def __init__(self, context):
    """
    :param context: The global pants :py:class:`twitter.pants.goal.context.Context`
    """
    self.context = context

    self._pants_workdir = self.context.config.getdefault('pants_workdir')
    default_cache_dir = self.get_workdir(section='build_cache', workdir='build_cache')
    self._build_cache = context.config.get('tasks', 'build_cache', default=default_cache_dir)
    self._basedir = os.path.join(self._build_cache, self.__class__.__name__)
    self._extradata = os.path.join(self._basedir, Task.EXTRA_DATA)

  def invalidate(self, all_=False):
    """
      Invalidates the target cache used by this task.

      The next check of changed is guaranteed to indicate that any target checked has in fact
      changed.

      :param bool all_: If True then all Task caches are invalidated, not just this tasks cache.
    """
    self.context.log.debug('Invalidating caches for %s' % 'all tasks' if all_ else str(self))
    safe_rmtree(self._build_cache if all_ else self._basedir)

  def execute(self, targets):
    """Executes this task against the given targets.

    :param targets: List of concrete :py:class:`twitter.pants.base.target.Target` instances,
      which may be a subset of the current context targets.
    """

  def invalidate_for(self):
    """
      Subclasses can override and return an object that should be checked for changes when using
      changed to manage target invalidation.  If the pickled form of returned object changes
      between runs all targets will be invalidated.
    """

  class CacheManager(object):
    """
      Manages cache checks, updates and invalidation keeping track of basic change and invalidation
      statistics.
    """
    def __init__(self, cache, targets, only_externaldeps):
      self._cache = cache
      self._targets = set(targets)
      self._sources = NO_SOURCES if only_externaldeps else TARGET_SOURCES

      self.changed_files = 0
      self.invalidated_files = 0
      self.invalidated_targets = 0
      self.foreign_invalidated_targets = 0
      self.changed = defaultdict(list)

    def check_content(self, identifier, files):
      """
        Checks if identified content has changed and invalidates it if so.

        :id An identifier for the tracked content.
        :files The files containing the content to track changes for.
        :returns: The cache key for this content.
      """
      cache_key = self._cache.key_for(identifier, files)
      if self._cache.needs_update(cache_key):
        return cache_key

    def check(self, target):
      """Checks if a target has changed and invalidates it if so."""
      cache_key = self._key_for(target)
      if cache_key and self._cache.needs_update(cache_key):
        self._invalidate(target, cache_key)

    def update(self, cache_key):
      """Mark a changed or invalidated target as successfully processed."""
      self._cache.update(cache_key)

    def invalidate(self, target, cache_key=None):
      """Forcefully mark a target as changed."""
      self._invalidate(target, cache_key or self._key_for(target), indirect=True)

    def _key_for(self, target):
      return self._cache.key_for_target(
        target,
        sources=self._sources,
        fingerprint_extra=lambda sha: self._fingerprint_jardeps(target, sha)
      )

    def _fingerprint_jardeps(self, target, sha):
      internaltargets = OrderedSet()
      alltargets = OrderedSet()
      def fingerprint_external(target):
        internaltargets.add(target)
        if hasattr(target, 'dependencies'):
          alltargets.update(target.dependencies)
      target.walk(fingerprint_external)

      for external_target in alltargets - internaltargets:
        sha.update((external_target.cache_key() if issubclass(external_target, ExternalDependency)
                    else str(external_target)).encode('utf8'))

    def _invalidate(self, target, cache_key, indirect=False):
      if target in self._targets:
        self.changed[target].append(cache_key)
        if indirect:
          self.invalidated_files += len(cache_key.sources)
          self.invalidated_targets += 1
        else:
          self.changed_files += len(cache_key.sources)
      else:
        # invalidate a target to be processed in a subsequent round - this handles goal groups
        self._cache.invalidate(cache_key)
        self.foreign_invalidated_targets += 1

  @contextmanager
  def changed(self, targets, only_buildfiles=False, invalidate_dependants=False):
    """
      Yields an iterable over the targets that have changed since the last check to a with block.
      If no exceptions are thrown by work in the block, the cache is updated for the targets,
      otherwise if a TargetError is thrown by the work in the block all targets except those in the
      TargetError are cached.

      :targets The targets to check for changes.
      :only_buildfiles If True, then just the target's BUILD files are checked for changes.
      :invalidate_dependants If True then any targets depending on changed targets are invalidated
      :returns: the subset of targets that have changed
    """

    safe_mkdir(self._basedir)
    cache_manager = Task.CacheManager(BuildCache(self._basedir), targets, only_buildfiles)

    check = self.invalidate_for()
    if check is not None:
      with safe_open(self._extradata, 'w') as pickled:
        pickle.dump(check, pickled)

      cache_key = cache_manager.check_content(Task.EXTRA_DATA, [self._extradata])
      if cache_key:
        self.context.log.debug('invalidating all targets for %s' % self.__class__.__name__)
        for target in targets:
          cache_manager.invalidate(target, cache_key)

    for target in targets:
      cache_manager.check(target)

    if invalidate_dependants and cache_manager.changed:
      for target in (self.context.dependants(lambda t: t in cache_manager.changed.keys())).keys():
        cache_manager.invalidate(target)

    if invalidate_dependants:
      if cache_manager.foreign_invalidated_targets:
        self.context.log.info('Invalidated %d dependant targets '
                              'for the next round' % cache_manager.foreign_invalidated_targets)

      if cache_manager.changed_files:
        msg = 'Operating on %d files in %d changed targets' % (
          cache_manager.changed_files,
          len(cache_manager.changed)
        )
        if cache_manager.invalidated_files:
          msg += ' and %d files in %d invalidated dependant targets' % (
            cache_manager.invalidated_files,
            cache_manager.invalidated_targets
          )
        self.context.log.info(msg)
    elif cache_manager.changed_files:
      self.context.log.info('Operating on %d files in %d changed targets' % (
        cache_manager.changed_files,
        len(cache_manager.changed)
      ))

    try:
      yield cache_manager.changed.keys()
      for cache_keys in cache_manager.changed.values():
        for cache_key in cache_keys:
          cache_manager.update(cache_key)
    except TargetError as e:
      for target, cache_keys in cache_manager.changed.items():
        if target not in e.targets:
          for cache_key in cache_keys:
            cache_manager.update(cache_key)

  def get_workdir(self, section="default", key="workdir", workdir=None):
    return self.context.config.get(section,
                                   key,
                                   default=os.path.join(self._pants_workdir, workdir or
                                                        self.__class__.__name__.lower()))

__all__ = (
  'TaskError',
  'TargetError',
  'Task'
)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from twitter.common.contextutil import pushd
from twitter.common.lang import Compatibility
from twitter.pants import get_buildroot
from twitter.pants.base import Target
from twitter.pants.targets.sources import SourceRoot


class TargetWithSources(Target):
  def __init__(self, name, sources=None):
    Target.__init__(self, name)

    self.target_base = SourceRoot.find(self)
    self.sources = None
    self._sources = sources or []

  def expand_files(self, recursive=True, include_buildfile=True):
    """Expand files used to build this target to absolute paths.  By default this expansion is done
    recursively and target BUILD files are included."""

    files = []

    def _expand(target):
      files.extend([os.path.abspath(os.path.join(target.target_base, s))
          for s in (target.sources or [])])
      if include_buildfile:
        files.append(target.address.buildfile.full_path)
      if recursive:
        for dep in target.dependencies:
          if isinstance(dep, TargetWithSources):
            _expand(dep)
          elif hasattr(dep, 'address'):
            # Don't know what it is, but we'll include the BUILD file to be paranoid
            files.append(dep.address.buildfile.full_path)

    _expand(self)
    return files

  @property
  def sources(self):
    if self._resolved_sources is None:
      self.sources = self._resolve_paths(self.target_base, self._sources or [])
    return self._resolved_sources

  @sources.setter
  def sources(self, sources):
    self._resolved_sources = sources

  def _resolve_paths(self, rel_base, paths):
    """
      Resolves paths relative to the given rel_base from the build root.
      For example:
        target: ~/workspace/src/java/com/twitter/common/base/BUILD
        rel_base: src/resources

      Resolves paths from:
        ~/workspace/src/resources/com/twitter/common/base
    """

    if not paths:
      return []

    def flatten_paths(*items):
      """Flattens one or more items into a list.  If the item is iterable each of its items is
      flattened.  If an item is callable, it is called and the result is flattened.  Otherwise the
      atom is appended to the flattened list.  These rules are applied recursively such that the
      returned list will only contain non-iterable, non-callable atoms."""

      flat = []

      def flatmap(item):
        if isinstance(item, Compatibility.string):
          flat.append(item)
        else:
          try:
            for i in iter(item):
              flatmap(i)
          except TypeError:
            if callable(item):
              flatmap(item())
            else:
              flat.append(item)

      for item in items:
        flatmap(item)

      return flat

    src_relpath = os.path.relpath(self.address.buildfile.parent_path,
                                  os.path.join(get_buildroot(), self.target_base))

    resolve_basepath = os.path.join(get_buildroot(), rel_base, src_relpath)
    with pushd(resolve_basepath):
      return [ os.path.normpath(os.path.join(src_relpath, path)) for path in flatten_paths(paths) ]



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.pants.base import manual
from twitter.pants.targets.exclude import Exclude

from .external_dependency import ExternalDependency


class Artifact(object):
  """
  Specification for an Ivy Artifact for this jar dependency.

  See: http://ant.apache.org/ivy/history/latest-milestone/ivyfile/artifact.html
  """
  def __init__(self, name, type_=None, ext=None, conf=None, url=None, classifier=None):
    """
    :param name: The name of the published artifact. This name must not include revision.
    :param type_: The type of the published artifact. It's usually its extension, but not
      necessarily. For instance, ivy files are of type 'ivy' but have 'xml' extension.
    :param ext: The extension of the published artifact.
    :param conf: The public configuration in which this artifact is published. The '*' wildcard can
      be used to designate all public configurations.
    :param url: The url at which this artifact can be found if it isn't located at the standard
      location in the repository
    :param classifier: The maven classifier of this artifact.
    """
    self.name = name
    self.type_ = type_ or 'jar'
    self.ext = ext
    self.url = url
    self.classifier = classifier
    self.conf = conf

  def __repr__(self):
    return ('Artifact(%r, type_=%r, ext=%r, conf=%r, url=%r, classifier=%r)'
            % (self.name, self.type_, self.ext, self.conf, self.url, self.classifier))


@manual.builddict(tags=["jvm"])
class  JarDependency(ExternalDependency):
  """A pre-built Maven repository dependency."""

  _JAR_HASH_KEYS = (
    'org',
    'name',
    'rev',
    'force',
    'excludes',
    'transitive',
    'ext',
    'url',
    'mutable',
    '_configurations'
  )

  def __init__(self, org, name, rev=None, force=False, ext=None, url=None, apidocs=None,
               type_=None, classifier=None, mutable=None):
    """
    :param string org: The Maven ``groupId`` of this dependency.
    :param string name: The Maven ``artifactId`` of this dependency.
    :param string rev: The Maven ``version`` of this dependency.
      If unspecified the latest available version is used.
    :param boolean force: Force this specific artifact revision even if other transitive
      dependencies specify a different revision. This requires specifying the ``rev`` parameter.
    :param string ext: Extension of the artifact if different from the artifact type.
      This is sometimes needed for artifacts packaged with Maven bundle type but stored as jars.
    :param string url: URL of this artifact, if different from the Maven repo standard location
      (specifying this parameter is unusual).
    :param string apidocs: URL of existing javadocs, which if specified, pants-generated javadocs
      will properly hyperlink {\ @link}s.
    :param string type_: Artifact packaging type.
    :param string classifier: Classifier specifying the artifact variant to use.
      Use ``with_artifact`` to include multiple artifacts with different classifiers.
    :param boolean mutable: Inhibit caching of this mutable artifact. A common use is for
      Maven -SNAPSHOT style artifacts in an active development/integration cycle.
    """
    self.org = org
    self.name = name
    self.rev = rev
    self.force = force
    self.excludes = []
    self.transitive = True
    self.apidocs = apidocs
    self.mutable = mutable
    self._classifier = classifier

    self.artifacts = []
    if ext or url or type_ or classifier:
      self.with_artifact(name=name, type_=type_, ext=ext, url=url, classifier=classifier)

    self.id = repr(self)
    self._configurations = ['default']

    # Support legacy method names
    # TODO(John Sirois): introduce a deprecation cycle for these and then kill
    self.withSources = self.with_sources
    self.withDocs = self.with_sources

  @property
  def classifier(self):
    """Returns the maven classifier for this jar dependency.

    If the classifier is ambiguous; ie: there was no classifier set in the constructor and the jar
    dependency has multiple attached artifacts, a :class:`ValueError` is raised.
    """
    if self._classifier or len(self.artifacts) == 0:
      return self._classifier
    elif len(self.artifacts) == 1:
      return self.artifacts[0].classifier
    else:
      raise ValueError('Cannot determine classifier. No explicit classifier is set and this jar '
                       'has more than 1 artifact: %s\n\t%s'
                       % (self, '\n\t'.join(map(str, self.artifacts))))

  @manual.builddict()
  def exclude(self, org, name=None):
    """Adds a transitive dependency of this jar to the exclude list."""

    self.excludes.append(Exclude(org, name))
    return self

  @manual.builddict()
  def intransitive(self):
    """Declares this Dependency intransitive, indicating only the jar for the dependency itself
    should be downloaded and placed on the classpath"""

    self.transitive = False
    return self

  @manual.builddict()
  def with_sources(self):
    """This requests the artifact have its source jar fetched.
    (This implies there *is* a source jar to fetch.) Used in contexts
    that can use source jars (as of 2013, just eclipse and idea goals)."""
    self._configurations.append('sources')
    return self

  def with_docs(self):
    self._configurations.append('docs')
    return self

  def with_artifact(self, name=None, type_=None, ext=None, url=None, configuration=None,
                    classifier=None):
    """Sets an alternative artifact to fetch or adds additional artifacts if called multiple times.
    """
    artifact = Artifact(name or self.name, type_=type_, ext=ext, url=url, conf=configuration,
                        classifier=classifier)
    self.artifacts.append(artifact)
    return self

  def __eq__(self, other):
    result = (isinstance(other, JarDependency)
              and self.org == other.org
              and self.name == other.name
              and self.rev == other.rev)
    return result

  def __hash__(self):
    return hash((self.org, self.name, self.rev))

  def __ne__(self, other):
    return not self.__eq__(other)

  def __repr__(self):
    return "%s-%s-%s" % (self.org, self.name, self.rev)

  def cache_key(self):
    return ''.join(getattr(self, key) for key in self._JAR_HASH_KEYS)

  def resolve(self):
    yield self

  def walk(self, work, predicate=None):
    if not predicate or predicate(self):
      work(self)

  def _as_jar_dependencies(self):
    yield self



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from twitter.common.lang import Compatibility
from twitter.common.log import logger
from twitter.pants.base.config import Config


log = logger(name='rcfile')


class RcFile(object):
  """Encapsulates pants rcfile handling."""

  # TODO(John Sirois): localize handling of this flag value back into pants_exe.py once the new old
  # split is healed.
  _DISABLE_PANTS_RC_OPTION = '--no-pantsrc'

  @staticmethod
  def install_disable_rc_option(parser):
    parser.add_option(RcFile._DISABLE_PANTS_RC_OPTION, action = 'store_true', dest = 'nopantsrc',
                      default = False, help = 'Specifies that pantsrc files should be ignored.')

  def __init__(self, paths, default_prepend=True, process_default=False):
    """
      Creates an RcFile object that gives precedence to rcfiles that come last in the given
      sequence of paths.  The effect is as if each rcfile in paths overlays the next in a walk from
      left to right.

      :paths The rcfiles to apply default subcommand options from.
      :default_prepend Whether to prepend (the default) or append if default options are specified
                       with the 'options' key.
      :process_default True to process options in the [DEFAULT] section and apply regardless of
                       goal.
    """

    self.default_prepend = default_prepend
    self.process_default = process_default

    if not paths:
      raise ValueError('One or more rcfile paths must be specified')

    if isinstance(paths, Compatibility.string):
      paths = [paths]
    self.paths = [os.path.expanduser(path) for path in paths]

  def apply_defaults(self, commands, args):
    """
      Augments the given list of arguments with any default options found for the given commands.
      The returned argments will be a new copy of the given args with possibly extra augmented
      arguments.

      Default options are applied from the following keys under a section with the name of the
      subcommand the default options apply to:
      'options': These options are either prepended or appended to the command line args as
                 specified in the constructor with default_prepend.
      'prepend-options': These options are prepended to the command line args.
      'append-options': These options are appended to the command line args.
    """

    args = args[:]

    if RcFile._DISABLE_PANTS_RC_OPTION in args:
      return args

    config = Config.create_parser()
    read_from = config.read(self.paths)
    if not read_from:
      log.debug('no rcfile found')
      return args

    log.debug('using rcfiles: %s to modify args' % ','.join(read_from))

    def get_rcopts(command, key):
      return config.get(command, key).split() if config.has_option(command, key) else []

    commands = list(commands)
    if self.process_default:
      commands.insert(0, Config.DEFAULT_SECTION)

    for cmd in commands:
      opts = get_rcopts(cmd, 'options')
      args = (opts + args) if self.default_prepend else (args + opts)
      args = get_rcopts(cmd, 'prepend-options') + args + get_rcopts(cmd, 'append-options')
    return args



# ==================================================================================================
# Copyright 2014 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.pants.targets.jvm_binary import JvmBinary
from twitter.pants.targets.exclude import Exclude

LOGGING_EXCLUDES = [
  Exclude(org = 'ch.qos.logback'),
  Exclude(org = 'commons-logging', name = 'commons-logging'),
  Exclude(org = 'log4j'),
  Exclude(org = 'org.slf4j'),
]
# The storm cluster provides the appropriate versions of these jars.
ON_STORM_CLUSTER_EXCLUDES = [
  Exclude(org = 'storm', name = 'storm'),
  Exclude(org = 'backtype', name = 'jzmq'),
  Exclude(org = 'ch.qos.logback', name = 'logback-classic'),
  Exclude(org = 'clj-time', name = 'clj-time'),
  Exclude(org = 'com.google.guava'),
  Exclude(org = 'com.googlecode.disruptor', name = 'disruptor'),
  Exclude(org = 'com.googlecode.json-simple', name = 'json-simple'),
  Exclude(org = 'com.netflix.curator', name = 'curator-framework'),
  Exclude(org = 'com.twitter.storm', name = 'carbonite'),
  Exclude(org = 'commons-io', name = 'commons-io'),
  Exclude(org = 'compojure', name = 'compojure'),
  Exclude(org = 'hiccup', name = 'hiccup'),
  Exclude(org = 'org.apache.commons', name = 'commons-exec'),
  Exclude(org = 'org.apache.httpcomponents', name = 'httpclient'),
  Exclude(org = 'org.clojure', name = 'clojure'),
  Exclude(org = 'org.clojure', name = 'math.numeric-tower'),
  Exclude(org = 'org.clojure', name = 'tools.logging'),
  Exclude(org = 'org.yaml', name = 'snakeyaml'),
  Exclude(org = 'ring', name = 'ring-jetty-adapter'),
  Exclude(org = 'storm', name = 'jgrapht'),
  Exclude(org = 'storm', name = 'libthrift7'),
  Exclude(org = 'storm', name = 'tools.cli'),
]

class TwitterStormBinary(JvmBinary):
  """A binary that is suitable for running on the Storm cluster.

  Invoking the ``binary`` or ``bundle`` goal on one of these targets creates a binary jar that
  excludes any dependencies already provided by the Storm cluster.
  """
  def __init__(
      self,
      name,
      main,
      basename=None,
      source=None,
      resources=None,
      dependencies=None,
      excludes=[],
      deploy_excludes=[]):

    JvmBinary.__init__(
      self,
      name=name,
      main=main,
      basename=basename,
      source=source,
      resources=resources,
      dependencies=dependencies,
      excludes=excludes,
      deploy_excludes=(deploy_excludes or []) + LOGGING_EXCLUDES + ON_STORM_CLUSTER_EXCLUDES,
      configurations=None
    )



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import functools
import os

from contextlib import contextmanager
from zipfile import ZIP_STORED, ZIP_DEFLATED

from twitter.common.dirutil import safe_mkdir

from twitter.pants import get_buildroot, has_resources, has_sources
from twitter.pants.fs import safe_filename
from twitter.pants.java.jar import open_jar
from twitter.pants.targets import JvmBinary, JavaThriftLibrary, ThriftLibrary
from twitter.pants.tasks import Task, TaskError
from twitter.pants.tasks.javadoc_gen import javadoc
from twitter.pants.tasks.scaladoc_gen import scaladoc

DEFAULT_CONFS = ['default']


def is_binary(target):
  return isinstance(target, JvmBinary)


def is_java_library(target):
  return has_sources(target, '.java') and not is_binary(target)


def is_scala_library(target):
  return has_sources(target, '.scala') and not is_binary(target)


def is_jvm_library(target):
  return is_java_library(target) or is_scala_library(target)


def is_idl_library(target):
  return has_sources(target, '.thrift') and isinstance(target, (ThriftLibrary, JavaThriftLibrary))


def jarname(target, extension='.jar'):
  # TODO(John Sirois): incorporate version
  _, id_, _ = target.get_artifact_info()
  # Cap jar names quite a bit lower than the standard fs limit of 255 characters since these
  # artifacts will often be used outside pants and those uses may manipulate (expand) the jar
  # filenames blindly.
  return safe_filename(id_, extension, max_length=200)


class JarCreate(Task):
  """Jars jvm libraries and optionally their sources and their docs."""

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("outdir"), dest="jar_create_outdir",
                            help="Emit jars in to this directory.")

    option_group.add_option(mkflag("compressed"), mkflag("compressed", negate=True),
                            dest="jar_create_compressed", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Create compressed jars.")

    option_group.add_option(mkflag("transitive"), mkflag("transitive", negate=True),
                            dest="jar_create_transitive", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Create jars for the transitive closure of internal "
                                 "targets reachable from the roots specified on the command line.")

    option_group.add_option(mkflag("classes"), mkflag("classes", negate=True),
                            dest="jar_create_classes", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Create class jars.")
    option_group.add_option(mkflag("sources"), mkflag("sources", negate=True),
                            dest="jar_create_sources", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Create source jars.")
    option_group.add_option(mkflag("javadoc"), mkflag("javadoc", negate=True),
                            dest="jar_create_javadoc", default=None,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Create javadoc jars (includes scaladoc etc.)")
    option_group.add_option(mkflag("idl"), mkflag("idl", negate=True),
                            dest="jar_create_idl", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Create Thrift jars.")

  def __init__(self, context, jar_javadoc=False):
    Task.__init__(self, context)

    options = context.options
    products = context.products

    self._output_dir = (options.jar_create_outdir or
                        self.get_workdir(section='jar-create', workdir='jars'))
    self.transitive = options.jar_create_transitive
    self.confs = context.config.getlist('jar-create', 'confs', default=DEFAULT_CONFS)
    self.compression = ZIP_DEFLATED if options.jar_create_compressed else ZIP_STORED

    self.jar_classes = options.jar_create_classes or products.isrequired('jars')
    if self.jar_classes:
      products.require('classes')

    self.jar_idl = products.isrequired('idl_jars') or options.jar_create_idl
    if self.jar_idl:
      products.require('idl')

    definitely_create_javadoc = options.jar_create_javadoc or products.isrequired('javadoc_jars')
    definitely_dont_create_javadoc = options.jar_create_javadoc is False
    create_javadoc = jar_javadoc and options.jar_create_javadoc is None
    if definitely_create_javadoc and definitely_dont_create_javadoc:
      self.context.log.warn('javadoc jars are required but you have requested they not be created, '
                            'creating anyway')
    self.jar_javadoc = (True  if definitely_create_javadoc      else
                        False if definitely_dont_create_javadoc else
                        create_javadoc)
    if self.jar_javadoc:
      products.require(javadoc.product_type)
      products.require(scaladoc.product_type)

    self.jar_sources = products.isrequired('source_jars') or options.jar_create_sources

    self._jars = {}

  def execute(self, targets):
    safe_mkdir(self._output_dir)

    def jar_targets(predicate):
      return filter(predicate, (targets if self.transitive else self.context.target_roots))

    def add_genjar(typename, target, name):
      self.context.products.get(typename).add(target, self._output_dir).append(name)

    if self.jar_classes:
      self.jar(jar_targets(is_jvm_library),
               self.context.products.get('classes'),
               functools.partial(add_genjar, 'jars'))

    if self.jar_idl:
      self.idljar(jar_targets(is_idl_library), functools.partial(add_genjar, 'idl_jars'))

    if self.jar_sources:
      self.sourcejar(jar_targets(is_jvm_library), functools.partial(add_genjar, 'source_jars'))

    if self.jar_javadoc:
      javadoc_add_genjar = functools.partial(add_genjar, 'javadoc_jars')
      self.javadocjar(jar_targets(is_java_library),
                      self.context.products.get(javadoc.product_type),
                      javadoc_add_genjar)
      self.javadocjar(jar_targets(is_scala_library),
                      self.context.products.get(scaladoc.product_type),
                      javadoc_add_genjar)

  @contextmanager
  def create_jar(self, target, path):
    existing = self._jars.setdefault(path, target)
    if target != existing:
      raise TaskError('Duplicate name: target %s tried to write %s already mapped to target %s' % (
        target, path, existing
      ))
    self._jars[path] = target
    with open_jar(path, 'w', compression=self.compression) as jar:
      yield jar

  def jar(self, jvm_targets, genmap, add_genjar):
    for target in jvm_targets:
      generated = genmap.get(target)
      if generated or has_resources(target):
        jar_name = jarname(target)
        add_genjar(target, jar_name)
        jar_path = os.path.join(self._output_dir, jar_name)
        with self.create_jar(target, jar_path) as zip:
          if generated:
            for basedir, classfiles in generated.items():
              for classfile in classfiles:
                zip.write(os.path.join(basedir, classfile), classfile)

          if has_resources(target):
            resources_genmap = self.context.products.get('resources')
            if resources_genmap:
              for resources in target.resources:
                resource_map = resources_genmap.get(resources)
                if resource_map:
                  for basedir, files in resource_map.items():
                    for resource in files:
                      zip.write(os.path.join(basedir, resource), resource)

  def idljar(self, idl_targets, add_genjar):
    for target in idl_targets:
      jar_name = jarname(target, '-idl.jar')
      add_genjar(target, jar_name)
      jar_path = os.path.join(self._output_dir, jar_name)
      with self.create_jar(target, jar_path) as zh:
        for source in target.sources:
          zh.write(os.path.join(get_buildroot(), target.target_base, source), source)

  def sourcejar(self, jvm_targets, add_genjar):
    for target in jvm_targets:
      jar_name = jarname(target, '-sources.jar')
      add_genjar(target, jar_name)
      jar_path = os.path.join(self._output_dir, jar_name)
      with self.create_jar(target, jar_path) as zip:
        for source in target.sources:
          zip.write(os.path.join(get_buildroot(), target.target_base, source), source)

        if has_resources(target):
          for resources in target.resources:
            for resource in resources.sources:
              zip.write(os.path.join(get_buildroot(), resources.target_base, resource), resource)

  def javadocjar(self, java_targets, genmap, add_genjar):
    for target in java_targets:
      generated = genmap.get(target)
      if generated:
        jar_name = jarname(target, '-javadoc.jar')
        add_genjar(target, jar_name)
        jar_path = os.path.join(self._output_dir, jar_name)
        with self.create_jar(target, jar_path) as zip:
          for basedir, javadocfiles in generated.items():
            for javadocfile in javadocfiles:
              zip.write(os.path.join(basedir, javadocfile), javadocfile)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from collections import namedtuple

import hashlib
import os
import re
import time

from twitter.common import log
from twitter.common.collections import maybe_list
from twitter.common.dirutil import safe_open
from twitter.common.lang import Compatibility

from twitter.pants import get_buildroot

from .executor import Executor, SubprocessExecutor

from . import NailgunClient


class NailgunExecutor(Executor):
  """Executes java programs by launching them in nailgun server.

  If a nailgun is not available for a given set of jvm args and classpath, one is launched and
  re-used for the given jvm args and classpath on subsequent runs.
  """

  class Endpoint(namedtuple('Endpoint', ['fingerprint', 'pid', 'port'])):
    """The coordinates for a nailgun server controlled by NailgunExecutor."""

    @classmethod
    def parse(cls, endpoint):
      """Parses an endpoint from a string of the form fingerprint:pid:port"""
      components = endpoint.split(':')
      if len(components) != 3:
        raise ValueError('Invalid endpoint spec %s' % endpoint)
      fingerprint, pid, port = components
      return cls(fingerprint, int(pid), int(port))

  # Used to identify we own a given java nailgun server
  PANTS_NG_ARG_PREFIX = '-Dtwitter.pants.buildroot'
  PANTS_NG_ARG = '%s=%s' % (PANTS_NG_ARG_PREFIX, get_buildroot())

  _PANTS_FINGERPRINT_ARG_PREFIX = '-Dfingerprint='

  @staticmethod
  def _check_pid(pid):
    try:
      os.kill(pid, 0)
      return True
    except OSError:
      return False

  @staticmethod
  def create_pidfile_arg(pidfile):
    return '-Dpidfile=%s' % os.path.relpath(pidfile, get_buildroot())

  @classmethod
  def _create_fingerprint_arg(cls, fingerprint):
    return cls._PANTS_FINGERPRINT_ARG_PREFIX + fingerprint

  @classmethod
  def parse_fingerprint_arg(cls, args):
    for arg in args:
      components = arg.split(cls._PANTS_FINGERPRINT_ARG_PREFIX)
      if len(components) == 2 and components[0] == '':
        return components[1]
    return None

  @staticmethod
  def _fingerprint(jvm_args, classpath):
    digest = hashlib.sha1()
    digest.update(''.join(sorted(jvm_args)))
    digest.update(''.join(sorted(classpath)))  # TODO(John Sirois): hash classpath contents?
    return digest.hexdigest()

  @staticmethod
  def _log_kill(pid, port=None, logger=None):
    logger = logger or log.info
    logger('killing ng server @ pid:%d%s' % (pid, ' port:%d' % port if port else ''))

  def __init__(self, workdir, nailgun_classpath, distribution=None, ins=None):
    super(NailgunExecutor, self).__init__(distribution=distribution)

    self._nailgun_classpath = maybe_list(nailgun_classpath)

    if not isinstance(workdir, Compatibility.string):
      raise ValueError('Workdir must be a path string, given %s' % workdir)

    self._pidfile = os.path.join(workdir, 'pid')
    self._ng_out = os.path.join(workdir, 'stdout')
    self._ng_err = os.path.join(workdir, 'stderr')

    self._ins = ins

  def _safe_execute(self, jvm_args, classpath, main, args):
    nailgun = self._get_nailgun_client(jvm_args, classpath)
    try:
      cmd = self._create_command(jvm_args, classpath, main, args)
      log.debug('Executing via %s: %s' % (nailgun, ' '.join(cmd)))
      return nailgun(main, *args)
    except nailgun.NailgunError as e:
      self.kill()
      raise self.Error('Problem launching via %s command %s %s: %s'
                       % (nailgun, main, ' '.join(args), e))

  def kill(self):
    """Kills the nailgun server owned by this executor if its currently running."""

    endpoint = self._get_nailgun_endpoint()
    if endpoint:
      self._log_kill(endpoint.pid, endpoint.port)
      try:
        os.kill(endpoint.pid, 9)
      except OSError:
        pass
      finally:
        os.remove(self._pidfile)

  def _get_nailgun_endpoint(self):
    if os.path.exists(self._pidfile):
      with safe_open(self._pidfile, 'r') as pidfile:
        contents = pidfile.read().strip()
        try:
          return self.Endpoint.parse(contents)
        except ValueError:
          log.warn('Invalid ng pidfile %s contained: %s' % (self._pidfile, contents))
          return None
    elif self._find:
      endpoint = self._find(self._pidfile)
      if endpoint:
        log.info('found ng server with fingerprint %s @ pid:%d port:%d' % endpoint)
        with safe_open(self._pidfile, 'w') as pidfile:
          pidfile.write('%s:%d:%d\n' % endpoint)
      return endpoint
    return None

  def _get_nailgun_client(self, jvm_args, classpath):
    classpath = self._nailgun_classpath + classpath
    new_fingerprint = self._fingerprint(jvm_args, classpath)

    endpoint = self._get_nailgun_endpoint()
    running = endpoint and self._check_pid(endpoint.pid)
    updated = endpoint and endpoint.fingerprint != new_fingerprint
    if running and not updated:
      return self._create_ngclient(port=endpoint.port)
    else:
      if running and updated:
        self.kill()
      return self._spawn_nailgun_server(new_fingerprint, jvm_args, classpath)

  # 'NGServer started on 127.0.0.1, port 53785.'
  _PARSE_NG_PORT = re.compile('.*\s+port\s+(\d+)\.$')

  def _parse_nailgun_port(self, line):
    match = self._PARSE_NG_PORT.match(line)
    if not match:
      raise NailgunClient.NailgunError('Failed to determine spawned ng port from response'
                                       ' line: %s' % line)
    return int(match.group(1))

  def _await_nailgun_server(self):
    nailgun_timeout_seconds = 5
    max_socket_connect_attempts = 10
    nailgun = None
    port_parse_start = time.time()
    with safe_open(self._ng_out, 'r') as ng_out:
      while not nailgun:
        started = ng_out.readline()
        if started:
          port = self._parse_nailgun_port(started)
          with open(self._pidfile, 'a') as pidfile:
            pidfile.write(':%d\n' % port)
          nailgun = self._create_ngclient(port)
          log.debug('Detected ng server up on port %d' % port)
        elif time.time() - port_parse_start > nailgun_timeout_seconds:
          raise NailgunClient.NailgunError('Failed to read ng output after'
                                           ' %s seconds' % nailgun_timeout_seconds)

    attempt = 0
    while nailgun:
      sock = nailgun.try_connect()
      if sock:
        sock.close()
        log.info('Connected to ng server with fingerprint'
                 ' %s pid: %d @ port: %d' % self._get_nailgun_endpoint())
        return nailgun
      elif attempt > max_socket_connect_attempts:
        raise nailgun.NailgunError('Failed to connect to ng output after %d connect attempts'
                                   % max_socket_connect_attempts)
      attempt += 1
      log.debug('Failed to connect on attempt %d' % attempt)
      time.sleep(0.1)

  def _create_ngclient(self, port):
    return NailgunClient(port=port, ins=self._ins, work_dir=get_buildroot())

  def _spawn_nailgun_server(self, fingerprint, jvm_args, classpath):
    log.info('No ng server found with fingerprint %s, spawning...' % fingerprint)

    with safe_open(self._ng_out, 'w'):
      pass  # truncate

    pid = os.fork()
    if pid != 0:
      # In the parent tine - block on ng being up for connections
      return self._await_nailgun_server()

    os.setsid()
    in_fd = open('/dev/null', 'r')
    out_fd = safe_open(self._ng_out, 'w')
    err_fd = safe_open(self._ng_err, 'w')

    java = SubprocessExecutor(self._distribution)

    jvm_args = jvm_args + [self.PANTS_NG_ARG,
                           self.create_pidfile_arg(self._pidfile),
                           self._create_fingerprint_arg(fingerprint)]

    process = java.spawn(jvm_args=jvm_args,
                         classpath=classpath,
                         main='com.martiansoftware.nailgun.NGServer',
                         args=[':0'],
                         stdin=in_fd,
                         stdout=out_fd,
                         stderr=err_fd,
                         close_fds=True,
                         cwd=get_buildroot())

    with safe_open(self._pidfile, 'w') as pidfile:
      pidfile.write('%s:%d' % (fingerprint, process.pid))
    log.debug('Spawned ng server with fingerprint %s @ %d' % (fingerprint, process.pid))
    # Prevents finally blocks and atexit handlers from being executed, unlike sys.exit(). We
    # don't want to execute finally blocks because we might, e.g., clean up tempfiles that the
    # parent still needs.
    os._exit(0)

  def __str__(self):
    return 'NailgunExecutor(%s, server=%s)' % (self._distribution, self._get_nailgun_endpoint())


try:
  import psutil

  def _find_ngs(everywhere=False):
    def cmdline_matches(cmdline):
      if everywhere:
        return any(filter(lambda arg: arg.startswith(NailgunExecutor.PANTS_NG_ARG_PREFIX), cmdline))
      else:
        return NailgunExecutor.PANTS_NG_ARG in cmdline

    for proc in psutil.process_iter():
      try:
        if 'java' == proc.name and cmdline_matches(proc.cmdline):
          yield proc
      except (psutil.AccessDenied, psutil.NoSuchProcess):
        pass

  def killall(logger=None, everywhere=False):
    success = True
    for proc in _find_ngs(everywhere=everywhere):
      try:
        NailgunExecutor._log_kill(proc.pid, logger=logger)
        proc.kill()
      except (psutil.AccessDenied, psutil.NoSuchProcess):
        success = False
    return success

  NailgunExecutor.killall = staticmethod(killall)

  def _find_ng_listen_port(proc):
    for connection in proc.get_connections(kind='tcp'):
      if connection.status == 'LISTEN':
        host, port = connection.laddr
        return port
    return None

  def _find(pidfile):
    pidfile_arg = NailgunExecutor.create_pidfile_arg(pidfile)
    for proc in _find_ngs(everywhere=False):
      try:
        if pidfile_arg in proc.cmdline:
          fingerprint = NailgunExecutor.parse_fingerprint_arg(proc.cmdline)
          port = _find_ng_listen_port(proc)
          if fingerprint and port:
            return NailgunExecutor.Endpoint(fingerprint, proc.pid, port)
      except (psutil.AccessDenied, psutil.NoSuchProcess):
        pass
    return None

  NailgunExecutor._find = staticmethod(_find)
except ImportError:
  NailgunExecutor.killall = None
  NailgunExecutor._find = None



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import re

from twitter.pants.base import manual, TargetDefinitionException

from .jar_dependency import JarDependency
from .java_thrift_library import JavaThriftLibrary
from .thrift_library import ThriftJar


@manual.builddict(tags=["java"])
class IdlJvmThriftLibrary(JavaThriftLibrary):
  """Creates a java thrift library from an idl jar

  Takes in an idl-only thrift jar and expands it to a JavaThriftLibrary with sources from the
  jar and the thrift_jar itself as the dependency.
  """
  # TODO (tina): keeping this around for now to be consistent with JavaThriftLibrary, but we need
  # to clean this up: https://jira.twitter.biz/browse/DPB-364
  _RPC_STYLE_DEFAULT = 'sync'

  def __init__(self,
               name,
               thrift_jar,
               language,
               excludes=None,
               rpc_style=_RPC_STYLE_DEFAULT,
               provided_by=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param thrift_jar: A :class:`twitter.pants.targets.thrift_library.ThriftJar` that contains
      the IDL files this target will generate sources from.
    :param excludes: List of :class:`twitter.pants.targets.exclude.Exclude` instances
      to filter this target's transitive dependencies against.
    :param language: The language used to generate the output files.
      One of 'java' or 'scala'.
    :param rpc_style: An optional rpc style to generate service stubs with.
      One of 'sync', 'finagle' or 'ostrich' with a default of 'sync'.
    :param provided_by: Published classfile jar if not default pants name.
    """
    JavaThriftLibrary.__init__(self, name=name, sources=[], provides=None,
                               dependencies=[thrift_jar], excludes=excludes, language=language,
                               rpc_style=rpc_style, compiler='scrooge')
    self.idl_jar = thrift_jar
    if not isinstance(thrift_jar, ThriftJar):
      raise TargetDefinitionException(
        self, "thrift_jar must be a ThriftJar instance, but was %s" % thrift_jar)

    # TODO (tina): Consolidate references to -only in a util https://jira.twitter.biz/browse/DPB-363
    provided_name = re.sub(r"-only$", "", thrift_jar.name)
    self.provided_by = provided_by or JarDependency(org=thrift_jar.org, name=provided_name,
                                                    rev=thrift_jar.rev).intransitive()
    if not isinstance(self.provided_by, JarDependency):
      raise TargetDefinitionException(self, "provided_by for target must be a JarDependency: %s"
                                            % self.provided_by)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.common.collections import OrderedSet
from twitter.pants.base import manual
from .ruby_target import RubyTarget


@manual.builddict(tags=["ruby"])
class RubyThriftLibrary(RubyTarget):
  """Generates a stub Ruby library from thrift IDL files."""

  def __init__(self, name,
               sources=None,
               dependencies=None,
               provides=None):
    """
    :param name: Name of library
    :param source: thrift source file
    :param dependencies: List of :class:`twitter.pants.base.target.Target` instances
      this target depends on.
    :type dependencies: list of targets
    """
    RubyTarget.__init__(self, name, sources, dependencies, provides)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

from twitter.pants import is_concrete, is_jvm, is_jvm_app, is_python, PythonRequirement
from twitter.pants.targets.jar_dependency import JarDependency

from . import TaskError
from .console_task import ConsoleTask


class Dependencies(ConsoleTask):
  """Generates a textual list (using the target format) for the dependency set of a target."""

  @staticmethod
  def _is_jvm(target):
    return is_jvm(target) or is_jvm_app(target)

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    super(Dependencies, cls).setup_parser(option_group, args, mkflag)

    cls.internal_only_flag = mkflag("internal-only")
    cls.external_only_flag = mkflag("external-only")

    option_group.add_option(cls.internal_only_flag,
                            action="store_true",
                            dest="dependencies_is_internal_only",
                            default=False,
                            help='Specifies that only internal dependencies should'
                                 ' be included in the graph output (no external jars).')
    option_group.add_option(cls.external_only_flag,
                            action="store_true",
                            dest="dependencies_is_external_only",
                            default=False,
                            help='Specifies that only external dependencies should'
                                 ' be included in the graph output (only external jars).')

  def __init__(self, context, **kwargs):
    super(Dependencies, self).__init__(context, **kwargs)

    if (self.context.options.dependencies_is_internal_only and
        self.context.options.dependencies_is_external_only):

      error_str = "At most one of %s or %s can be selected." % (self.internal_only_flag,
                                                                self.external_only_flag)
      raise TaskError(error_str)

    self.is_internal_only = self.context.options.dependencies_is_internal_only
    self.is_external_only = self.context.options.dependencies_is_external_only

  def console_output(self, unused_method_argument):
    for target in self.context.target_roots:
      if all(self._is_jvm(t) for t in target.resolve() if is_concrete(t)):
        for line in self._dependencies_list(target):
          yield line

      elif is_python(target):
        if self.is_internal_only:
          raise TaskError('Unsupported option for Python target: is_internal_only: %s' %
                          self.is_internal_only)
        if self.is_external_only:
          raise TaskError('Unsupported option for Python target: is_external_only: %s' %
                          self.is_external_only)
        for line in self._python_dependencies_list(target):
          yield line

  def _dep_id(self, dep):
    if isinstance(dep, JarDependency):
      if dep.rev:
        return False, '%s:%s:%s' % (dep.org, dep.name, dep.rev)
      else:
        return True, '%s:%s' % (dep.org, dep.name)
    else:
      return True, str(dep.address)

  def _python_dependencies_list(self, target):
    if isinstance(target, PythonRequirement):
      yield str(target._requirement)
    else:
      yield str(target.address)

    if hasattr(target, 'dependencies'):
      for dep in target.dependencies:
        for d in dep.resolve():
          for dep in self._python_dependencies_list(d):
            yield dep

  def _dependencies_list(self, target):
    def print_deps(visited, dep):
      internal, address = self._dep_id(dep)

      if not dep in visited:
        if internal and (not self.is_external_only or self.is_internal_only):
          yield address

        visited.add(dep)

        if self._is_jvm(dep):
          for internal_dependency in dep.internal_dependencies:
            for line in print_deps(visited, internal_dependency):
              yield line

        if not self.is_internal_only:
          if self._is_jvm(dep):
            for jar_dep in dep.jar_dependencies:
              internal, address  = self._dep_id(jar_dep)
              if not internal:
                if jar_dep not in visited:
                  if self.is_external_only or not self.is_internal_only:
                    yield address
                  visited.add(jar_dep)

    visited = set()
    for t in target.resolve():
      for dep in print_deps(visited, t):
        yield dep



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import marshal
import os
import re

from glob import glob1

from twitter.common.collections import OrderedSet
from twitter.common.python.interpreter import PythonIdentity

class BuildFile(object):
  _CANONICAL_NAME = 'BUILD'
  _PATTERN = re.compile('^%s(\.[a-z]+)?$' % _CANONICAL_NAME)

  @staticmethod
  def _is_buildfile_name(name):
    return BuildFile._PATTERN.match(name)

  @staticmethod
  def scan_buildfiles(root_dir, base_path=None):
    """Looks for all BUILD files under base_path"""

    buildfiles = []
    for root, dirs, files in os.walk(base_path if base_path else root_dir):
      for filename in files:
        if BuildFile._is_buildfile_name(filename):
          buildfile_relpath = os.path.relpath(os.path.join(root, filename), root_dir)
          buildfiles.append(BuildFile(root_dir, buildfile_relpath))
    return OrderedSet(sorted(buildfiles, key=lambda buildfile: buildfile.full_path))

  def __init__(self, root_dir, relpath, must_exist=True):
    """Creates a BuildFile object representing the BUILD file set at the specified path.

    root_dir: The base directory of the project
    relpath: The path relative to root_dir where the BUILD file is found - this can either point
        directly at the BUILD file or else to a directory which contains BUILD files
    must_exist: If True, the specified BUILD file must exist or else an IOError is thrown
    raises IOError if the specified path does not house a BUILD file and must_exist is True
    """

    path = os.path.abspath(os.path.join(root_dir, relpath))
    buildfile = os.path.join(path, BuildFile._CANONICAL_NAME) if os.path.isdir(path) else path

    if os.path.isdir(buildfile):
      raise IOError("%s is a directory" % buildfile)

    if must_exist:
      if not os.path.exists(buildfile):
        raise IOError("BUILD file does not exist at: %s" % buildfile)

      if not BuildFile._is_buildfile_name(os.path.basename(buildfile)):
        raise IOError("%s is not a BUILD file" % buildfile)

      if not os.path.exists(buildfile):
        raise IOError("BUILD file does not exist at: %s" % buildfile)

    self.root_dir = os.path.realpath(root_dir)
    self.full_path = os.path.realpath(buildfile)

    self.name = os.path.basename(self.full_path)
    self.parent_path = os.path.dirname(self.full_path)

    self._bytecode_path = os.path.join(self.parent_path, '.%s.%s.pyc' % (
      self.name, PythonIdentity.get()))

    self.relpath = os.path.relpath(self.full_path, self.root_dir)
    self.canonical_relpath = os.path.join(os.path.dirname(self.relpath), BuildFile._CANONICAL_NAME)

  def exists(self):
    """Returns True if this BuildFile corresponds to a real BUILD file on disk."""
    return os.path.exists(self.full_path)

  def descendants(self):
    """Returns all BUILD files in descendant directories of this BUILD file's parent directory."""

    descendants = BuildFile.scan_buildfiles(self.root_dir, self.parent_path)
    for sibling in self.family():
      descendants.discard(sibling)
    return descendants

  def ancestors(self):
    """Returns all BUILD files in ancestor directories of this BUILD file's parent directory."""

    def find_parent(dir):
      parent = os.path.dirname(dir)
      buildfile = os.path.join(parent, BuildFile._CANONICAL_NAME)
      if os.path.exists(buildfile) and not os.path.isdir(buildfile):
        return parent, BuildFile(self.root_dir, os.path.relpath(buildfile, self.root_dir))
      else:
        return parent, None

    parent_buildfiles = OrderedSet()

    parentdir = os.path.dirname(self.full_path)
    visited = set()
    while parentdir not in visited and self.root_dir != parentdir:
      visited.add(parentdir)
      parentdir, buildfile = find_parent(parentdir)
      if buildfile:
        parent_buildfiles.update(buildfile.family())

    return parent_buildfiles

  def siblings(self):
    """Returns an iterator over all the BUILD files co-located with this BUILD file not including
    this BUILD file itself"""

    for build in glob1(self.parent_path, 'BUILD*'):
      if self.name != build and BuildFile._is_buildfile_name(build):
        siblingpath = os.path.join(os.path.dirname(self.relpath), build)
        if not os.path.isdir(os.path.join(self.root_dir, siblingpath)):
          yield BuildFile(self.root_dir, siblingpath)

  def family(self):
    """Returns an iterator over all the BUILD files co-located with this BUILD file including this
    BUILD file itself.  The family forms a single logical BUILD file composed of the canonical BUILD
    file and optional sibling build files each with their own extension, eg: BUILD.extras."""

    yield self
    for sibling in self.siblings():
      yield sibling

  def code(self):
    """Returns the code object for this BUILD file."""
    if (os.path.exists(self._bytecode_path)
        and os.path.getmtime(self.full_path) <= os.path.getmtime(self._bytecode_path)):
      with open(self._bytecode_path, 'rb') as bytecode:
        return marshal.load(bytecode)
    else:
      with open(self.full_path, 'rb') as source:
        code = compile(source.read(), self.full_path, 'exec')
        with open(self._bytecode_path, 'wb') as bytecode:
          marshal.dump(code, bytecode)
        return code

  def __eq__(self, other):
    result = other and (
      type(other) == BuildFile) and (
      self.full_path == other.full_path)
    return result

  def __hash__(self):
    return hash(self.full_path)

  def __ne__(self, other):
    return not self.__eq__(other)

  def __repr__(self):
    return self.relpath



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from .jvm_target import JvmTarget
from .resources import Resources
from twitter.pants.base import manual


@manual.builddict(tags=["jvm"])
class Benchmark(JvmTarget):
  """A caliper benchmark.

  Run it with the ``bench`` goal.
  """

  def __init__(self,
               name,
               sources = None,
               java_sources = None,
               dependencies = None,
               excludes = None,
               resources = None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param sources: A list of filenames representing the source code
      this library is compiled from.
    :type sources: list of strings
    :param java_sources:
      :class:`twitter.pants.targets.java_library.JavaLibrary` or list of
      JavaLibrary targets this library has a circular dependency on.
      Prefer using dependencies to express non-circular dependencies.
    :param dependencies: List of :class:`twitter.pants.base.target.Target` instances
      this target depends on.
    :type dependencies: list of targets
    :param excludes: List of :class:`twitter.pants.targets.exclude.Exclude` instances
      to filter this target's transitive dependencies against.
    :param resources: An optional list of ``resources`` targets containing text
      file resources to place in this module's jar.
    """

    JvmTarget.__init__(self,
                       name,
                       sources,
                       dependencies,
                       excludes)

    self.java_sources = java_sources
    self.resources = list(self.resolve_all(resources, Resources))



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os
import shutil

from twitter.pants.ivy import profile_classpath
from twitter.pants.java import runjava

from .jvm_task import JvmTask

from . import Task, TaskError


class BenchmarkRun(JvmTask):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("target"), dest="target_class", action="append",
                            help="Name of the benchmark class.")

    option_group.add_option(mkflag("memory"), mkflag("memory", negate=True),
                            dest="memory_profiling", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Enable memory profiling.")

    option_group.add_option(mkflag("debug"), mkflag("debug", negate=True),
                            dest="debug", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Enable caliper debug mode.")

    option_group.add_option(mkflag("caliper-args"), dest="extra_caliper_args", default=[],
                            action="append",
                            help="Allows the user to pass additional command line options to "
                                 "caliper. Can be used multiple times and arguments will be "
                                 "concatenated. Example use: --bench-caliper-args='-Dsize=10,20 "
                                 "-Dcomplex=true,false' --bench-caliper-args=-Dmem=1,2,3")

  def __init__(self, context):
    Task.__init__(self, context)
    self.profile = context.config.get('benchmark-run', 'profile',
                                      default="benchmark-caliper-0.5")
    self.confs = context.config.getlist('benchmark-run', 'confs')
    self.jvm_args = context.config.getlist('benchmark-run', 'jvm_args',
                                            default=['-Xmx1g', '-XX:MaxPermSize=256m'])
    self.agent_profile = context.config.get('benchmark-run', 'agent_profile',
                                            default="benchmark-java-allocation-instrumenter-2.1")
    # TODO(Steve Gury):
    # Find all the target classes from the Benchmark target itself
    # https://jira.twitter.biz/browse/AWESOME-1938
    self.caliper_args = context.options.target_class

    if context.options.memory_profiling:
      self.caliper_args += ['--measureMemory']
      # For rewriting JDK classes to work, the JAR file has to be listed specifically in
      # the JAR manifest as something that goes in the bootclasspath.
      # The MANIFEST list a jar 'allocation.jar' this is why we have to rename it
      updated, classpath = profile_classpath(self.agent_profile)
      agent_jar = os.readlink(classpath[0])
      allocation_jar = os.path.join(os.path.dirname(agent_jar), "allocation.jar")
      if updated:
        shutil.copyfile(agent_jar, allocation_jar)
      os.environ['ALLOCATION_JAR'] = str(allocation_jar)

    if context.options.debug:
      self.jvm_args.extend(context.config.getlist('jvm', 'debug_args'))
      self.caliper_args += ['--debug']

    self.caliper_args.extend(context.options.extra_caliper_args)

  def execute(self, targets):
    caliper_main = 'com.google.caliper.Runner'
    _, classpath = profile_classpath(self.profile)
    exit_code = runjava(self.classpath(classpath, confs=self.confs),
                        caliper_main,
                        args=self.caliper_args,
                        jvm_args=self.jvm_args)
    if exit_code != 0:
      raise TaskError('java %s ... exited non-zero (%i)' % (caliper_main, exit_code))



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import json

from thrift.Thrift import TType

class ThriftJSONDecoder(json.JSONDecoder):
  """A decoder that makes python thrift structs JSON deserializable via the
  standard python json module.

  Pass this decoder when reading json, like this:

  json.loads(str, cls=text.ThriftJSONDecoder, <other kwargs>)

  Note that this is not a full protocol implementation in the thrift sense. This
  is just a quick-and-easy parser for unittests etc.
  """
  ROOT_THRIFT_CLASS = 'root_thrift_class'

  def __init__(self, *args, **kwargs):
    self.root_thrift_class = kwargs[ThriftJSONDecoder.ROOT_THRIFT_CLASS]
    del kwargs[ThriftJSONDecoder.ROOT_THRIFT_CLASS]
    super(ThriftJSONDecoder, self).__init__(*args, **kwargs)

  def decode(self, json_str):
    dict = super(ThriftJSONDecoder, self).decode(json_str)
    return self._convert(dict, TType.STRUCT,
                         (self.root_thrift_class, self.root_thrift_class.thrift_spec))

  def _convert(self, val, ttype, ttype_info):
    if ttype == TType.STRUCT:
      (thrift_class, thrift_spec) = ttype_info
      ret = thrift_class()
      for field in thrift_spec:
        if field is not None:
          (tag, field_ttype, field_name, field_ttype_info, dummy) = field
          if field_name in val:
            converted_val = self._convert(val[field_name], field_ttype, field_ttype_info)
            setattr(ret, field_name, converted_val)
    elif ttype == TType.LIST:
      (element_ttype, element_ttype_info) = ttype_info
      ret = [self._convert(x, element_ttype, element_ttype_info) for x in val]
    elif ttype == TType.SET:
      (element_ttype, element_ttype_info) = ttype_info
      ret = set([self._convert(x, element_ttype, element_ttype_info) for x in val])
    elif ttype == TType.MAP:
      (key_ttype, key_ttype_info, val_ttype, val_ttype_info) = ttype_info
      ret = dict([(self._convert(k, key_ttype, key_ttype_info),
                   self._convert(v, val_ttype, val_ttype_info)) for (k, v) in val.iteritems()])
    elif ttype == TType.STRING:
      ret = unicode(val)
    elif ttype == TType.DOUBLE:
      ret = float(val)
    elif ttype == TType.I64:
      ret = long(val)
    elif ttype == TType.I32 or ttype == TType.I16 or ttype == TType.BYTE:
      ret = int(val)
    elif ttype == TType.BOOL:
      ret = not not val
    else:
      raise Exception, 'Unrecognized thrift field type: %d' % ttype
    return ret

def json_to_thrift(json_str, root_thrift_class):
  """A utility shortcut function to parse a thrift json object of the specified class."""
  return json.loads(json_str, cls=ThriftJSONDecoder, root_thrift_class=root_thrift_class)




# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

from collections import defaultdict

from twitter.common.collections import OrderedDict, OrderedSet
from twitter.pants.base import manual
from twitter.pants.goal import GoalError


class SingletonPhases(type):
  phases = dict()
  renames = dict()

  def rename(cls, phase, name):
    """
      Renames the given phase and ensures all future requests for the old name are mapped to the
      given phase instance.
    """
    cls.phases.pop(phase.name)
    cls.renames[phase.name] = name
    phase.name = name
    cls.phases[name] = phase

  def __call__(cls, name):
    name = cls.renames.get(name, name)
    if name not in cls.phases:
      cls.phases[name] = super(SingletonPhases, cls).__call__(name)
    return cls.phases[name]

# Python 2.x + 3.x wankery
PhaseBase = SingletonPhases('PhaseBase', (object,), {})


@manual.builddict()
class Phase(PhaseBase):
  _goals_by_phase = defaultdict(list)
  _phase_by_goal = dict()

  @staticmethod
  def of(goal):
    return Phase._phase_by_goal[goal]

  @staticmethod
  def setup_parser(parser, args, phases):
    def do_setup_parser(phase, setup):
      for goal in phase.goals():
        if goal not in setup:
          setup.add(goal)
          for dep in goal.dependencies:
            do_setup_parser(dep, setup)
          goal.setup_parser(phase, parser, args)

    setup = set()
    for phase in phases:
      do_setup_parser(phase, setup)

  @staticmethod
  def all():
    """Returns all registered goals as a sorted sequence of phase, goals tuples."""
    return sorted(Phase._goals_by_phase.items(), key=lambda pair: pair[0].name)

  def __init__(self, name):
    self.name = name
    self.description = None

  def with_description(self, description):
    self.description = description
    return self

  def install(self, goal, first=False, replace=False, before=None, after=None):
    """
      Installs the given goal in this phase.  The placement of the goal in this phases' execution
      list defaults to the end but its position can be influence by specifying exactly one of the
      following arguments:

      first: Places the goal 1st in the execution list
      replace: Removes all existing goals in this phase and installs this goal
      before: Places the goal before the named goal in the execution list
      after: Places the goal after the named goal in the execution list
    """

    if (first or replace or before or after) and not (first ^ replace ^ bool(before) ^ bool(after)):
      raise GoalError('Can only specify one of first, replace, before or after')

    Phase._phase_by_goal[goal] = self

    g = self.goals()
    if replace:
      del g[:]
    g_names = map(lambda goal: goal.name, g)
    if first:
      g.insert(0, goal)
    elif before in g_names:
      g.insert(g_names.index(before), goal)
    elif after in g_names:
      g.insert(g_names.index(after) + 1, goal)
    else:
      g.append(goal)
    return self

  def rename(self, name):
    """Renames this goal."""
    PhaseBase.rename(self, name)
    return self

  def copy_to(self, name):
    """Copies this phase to the new named phase carrying along goal dependencies and description."""
    copy = Phase(name)
    copy.goals().extend(self.goals())
    copy.description = self.description
    return copy

  def remove(self, name):
    """Removes the named goal from this phase's list of goals to attempt."""
    goals = self.goals()
    for goal in goals:
      if goal.name == name:
        goals.remove(goal)
        return self
    raise GoalError('Goal %s does not exist in this phase, members are: %s' % (name, goals))

  class UnsatisfiedDependencyError(GoalError):
    """Raised when an operation cannot be completed due to an unsatisfied goal dependency."""

  def uninstall(self):
    """
      Removes the named phase and all its attached goals.  Raises Phase.UnsatisfiedDependencyError
      if the removal cannot be completed due to a dependency.
    """
    for phase, goals in Phase._goals_by_phase.items():
      for goal in goals:
        for dependee_phase in goal.dependencies:
          if self is dependee_phase:
            raise Phase.UnsatisfiedDependencyError(
              '%s is depended on by %s:%s' % (self.name, phase.name, goal.name))
    del Phase._goals_by_phase[self]

  def goals(self):
    return Phase._goals_by_phase[self]

  def __repr__(self):
    return self.name



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os
import textwrap

from collections import defaultdict

from twitter.common.dirutil import safe_mkdir, safe_open

from twitter.pants import has_sources, is_scalac_plugin
from twitter.pants.targets import resolve_target_sources, ScalaLibrary, ScalaTests
from twitter.pants.tasks import TaskError
from twitter.pants.tasks.jvm_compiler_dependencies import Dependencies
from twitter.pants.tasks.nailgun_task import NailgunTask


# Well known metadata file required to register scalac plugins with nsc.
_PLUGIN_INFO_FILE = 'scalac-plugin.xml'


def _is_scala(target):
  return has_sources(target, '.scala')


class ScalaCompile(NailgunTask):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    NailgunTask.setup_parser(option_group, args, mkflag)

    option_group.add_option(mkflag("warnings"), mkflag("warnings", negate=True),
                            dest="scala_compile_warnings", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Compile scala code with all configured warnings "
                                 "enabled.")

  def __init__(self, context):
    super(ScalaCompile, self).__init__(context)

    self._compile_profile = context.config.get('scala-compile', 'compile-profile')
    self._depemitter_profile = context.config.get('scala-compile', 'dependencies-plugin-profile')

    # All scala targets implicitly depend on the selected scala runtime.
    scaladeps = []
    for spec in context.config.getlist('scala-compile', 'scaladeps'):
      scaladeps.extend(context.resolve(spec))
    for target in context.targets(_is_scala):
      target.update_dependencies(scaladeps)

    workdir = context.config.get('scala-compile', 'workdir')
    self._classes_dir = os.path.join(workdir, 'classes')

    self._main = context.config.get('scala-compile', 'main')

    self._jvm_args = context.config.getlist('scala-compile', 'jvm_args', default=[])

    self._opts = context.config.getlist('scala-compile', 'args')
    if context.options.scala_compile_warnings:
      self._opts.extend(context.config.getlist('scala-compile', 'warning_args'))
    else:
      self._opts.extend(context.config.getlist('scala-compile', 'no_warning_args'))

    self._confs = context.config.getlist('scala-compile', 'confs')
    self._depfile = os.path.join(workdir, 'dependencies')

    # If we are compiling scala libraries with circular deps on java libraries we need to make sure
    # those cycle deps are present.
    self._inject_java_cycles()

  def _inject_java_cycles(self):
    for scala_target in self.context.targets(lambda t: isinstance(t, ScalaLibrary)):
      for java_target in scala_target.java_sources:
        self.context.add_target(java_target)

  def execute(self, targets):
    scala_targets = filter(_is_scala, targets)
    if scala_targets:
      compiler_classpath = self.profile_classpath(self._compile_profile)
      depemitter_classpath = self.profile_classpath(self._depemitter_profile)

      with self.context.state('classpath', []) as cp:
        for conf in self._confs:
          cp.insert(0, (conf, self._classes_dir))

      with self.changed(scala_targets, invalidate_dependants=True) as changed_targets:
        sources_by_target = self.calculate_sources(changed_targets)
        if sources_by_target:
          sources = reduce(lambda all, sources: all.union(sources), sources_by_target.values())
          if not sources:
            self.context.log.warn('Skipping scala compile for targets with no sources:\n  %s' %
                                  '\n  '.join(str(t) for t in sources_by_target.keys()))
          else:
            classpath = [jar for conf, jar in cp if conf in self._confs]
            result = self.compile(compiler_classpath, depemitter_classpath, classpath, sources)
            if result != 0:
              raise TaskError('java %s ... exited non-zero (%i)' % (self._main, result))

      if self.context.products.isrequired('classes'):
        genmap = self.context.products.get('classes')

        # Map generated classes to the owning targets and sources.
        dependencies = Dependencies(self._classes_dir, self._depfile)
        for target, classes_by_source in dependencies.findclasses(targets).items():
          for source, classes in classes_by_source.items():
            genmap.add(source, self._classes_dir, classes)
            genmap.add(target, self._classes_dir, classes)

        # Create and Map scala plugin info files to the owning targets.
        for target in targets:
          if is_scalac_plugin(target) and target.classname:
            basedir = self.write_plugin_info(target)
            genmap.add(target, basedir, [_PLUGIN_INFO_FILE])

  def calculate_sources(self, targets):
    sources = defaultdict(set)
    def collect_sources(target):
      src = (os.path.join(target.target_base, source)
             for source in target.sources if source.endswith('.scala'))
      if src:
        sources[target].update(src)

        if (isinstance(target, ScalaLibrary) or isinstance(target, ScalaTests)) and (
            target.java_sources):
          sources[target].update(resolve_target_sources(target.java_sources, '.java'))

    for target in targets:
      collect_sources(target)
    return sources

  def compile(self, compiler_classpath, depemitter_classpath, classpath, sources):
    safe_mkdir(self._classes_dir)

    # TODO(John Sirois): separate compiler profile from runtime profile
    args = [
      '-classpath', ':'.join(compiler_classpath + classpath),
      '-d', self._classes_dir,

      # Support for outputting a dependencies file of source -> class
      '-Xplugin:%s' % self.get_depemitter_plugin(depemitter_classpath),
      '-P:depemitter:file:%s' % self._depfile
    ]

    args.extend(self._opts)
    args.extend(sources)
    return self.runjava(compiler_classpath, self._main, args=args, jvm_args=self._jvm_args)

  def get_depemitter_plugin(self, depemitter_classpath):
    classpath = list(depemitter_classpath)
    depemitter_jar = classpath.pop()
    if classpath:
      raise TaskError('Expected only 1 jar for the depemitter plugin, '
                      'found these extra: ' % classpath)
    return depemitter_jar

  def write_plugin_info(self, target):
    basedir = os.path.join(self._classes_dir, target.id)
    with safe_open(os.path.join(basedir, _PLUGIN_INFO_FILE), 'w') as f:
      f.write(textwrap.dedent('''
        <plugin>
          <name>%s</name>
          <classname>%s</classname>
        </plugin>
      ''' % (target.plugin, target.classname)).strip())
    return basedir



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = 'John Sirois'

import os
import shutil

from twitter.common.dirutil import safe_mkdir

from twitter.pants import has_resources
from twitter.pants.tasks import Task


class PrepareResources(Task):

  def __init__(self, context):
    Task.__init__(self, context)

    self.workdir = context.config.get('prepare-resources', 'workdir')
    self.confs = context.config.getlist('prepare-resources', 'confs')

  def execute(self, targets):
    def extract_resources(target):
      return target.resources if has_resources(target) else ()
    all_resources = set()
    for resources in map(extract_resources, targets):
      all_resources.update(resources)

    def target_dir(resources):
      return os.path.join(self.workdir, resources.id)

    with self.changed(all_resources) as changed:
      for resources in changed:
        resources_dir = target_dir(resources)
        safe_mkdir(resources_dir, clean=True)
        for resource in resources.sources:
          basedir = os.path.dirname(resource)
          destdir = os.path.join(resources_dir, basedir)
          safe_mkdir(destdir)
          shutil.copy(
            os.path.join(resources.target_base, resource),
            os.path.join(resources_dir, resource)
          )

    genmap = self.context.products.get('resources')
    with self.context.state('classpath', []) as cp:
      for resources in all_resources:
        resources_dir = target_dir(resources)
        genmap.add(resources, resources_dir, resources.sources)
        for conf in self.confs:
          cp.insert(0, (conf, resources_dir))



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from contextlib import closing

from twitter.common.lang import Compatibility
StringIO = Compatibility.StringIO


class Manifest(object):
  """
    Implements the basics of the jar manifest specification.

    See: http://docs.oracle.com/javase/1.5.0/docs/guide/jar/jar.html#Manifest Specification
  """
  @staticmethod
  def _wrap(text):
    with closing(StringIO(text)) as fp:
      yield fp.read(70)
      while True:
        chunk = fp.read(69)
        if not chunk:
          return
        yield ' %s' % chunk

  PATH = 'META-INF/MANIFEST.MF'

  MANIFEST_VERSION = 'Manifest-Version'
  CREATED_BY = 'Created-By'
  MAIN_CLASS = 'Main-Class'
  CLASS_PATH = 'Class-Path'

  def __init__(self, contents=''):
    self._contents = contents.strip()

  def addentry(self, header, value):
    if len(header) > 68:
      raise ValueError('Header name must be 68 characters or less, given %s' % header)
    if self._contents:
      self._contents += '\n'
    self._contents += '\n'.join(self._wrap('%s: %s' % (header, value)))

  def contents(self):
    return self._contents + '\n'



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from functools import partial

from twitter.common.collections import maybe_list, OrderedSet

from twitter.pants.base import manual, Target
from twitter.pants.base.target import TargetDefinitionException
from twitter.pants.targets.anonymous import AnonymousDeps
from twitter.pants.targets.exclude import Exclude
from twitter.pants.targets.external_dependency import ExternalDependency
from twitter.pants.targets.exportable_jvm_library import ExportableJvmLibrary
from twitter.pants.targets.pants_target import Pants
from twitter.pants.targets.jar_dependency import JarDependency


@manual.builddict(tags=["anylang"])
class JarLibrary(Target):
  """A set of dependencies that may be depended upon,
  as if depending upon the set of dependencies directly.
  """

  def __init__(self, name, dependencies, overrides=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param dependencies: List of :class:`twitter.pants.base.target.Target` instances
      this target depends on.
    :param overrides: List of strings, each of which will be recursively resolved to
      any targets that provide artifacts. Those artifacts will override corresponding
      direct/transitive dependencies in the dependencies list.
    """
    Target.__init__(self, name)
    self._pre_override_dependencies = OrderedSet(maybe_list(dependencies,
      expected_type=(ExternalDependency, AnonymousDeps, Target),
      raise_type=partial(TargetDefinitionException, self)))
    self._dependencies = None
    self.override_targets = set(map(Pants, overrides or []))

  @property
  def dependencies(self):
    if not self._dependencies:
      # compute overridden dependencies
      self._dependencies = self._resolve_overrides()
    return self._dependencies

  def resolve(self):
    yield self
    for dependency in self.dependencies:
      for resolved_dependency in dependency.resolve():
        yield resolved_dependency

  def _resolve_overrides(self):
    """
    Resolves override targets, and then excludes and re-includes each of them
    to create and return a new dependency set.
    """
    if not self.override_targets:
      return self._pre_override_dependencies

    result = OrderedSet()

    # resolve overrides and fetch all of their "artifact-providing" dependencies
    excludes = set()
    for override_target in self.override_targets:
      # add pre_override deps of the target as exclusions
      for resolved in override_target.resolve():
        excludes.update(self._excludes(resolved))
      # prepend the target as a new target
      result.add(override_target)

    # add excludes for each artifact
    for direct_dep in self._pre_override_dependencies:
      # add relevant excludes to jar dependencies
      for jar_dep in self._jar_dependencies(direct_dep):
        for exclude in excludes:
          jar_dep.exclude(exclude.org, exclude.name)
      result.add(direct_dep)

    return result

  def _excludes(self, dep):
    """
    A generator for Exclude objects that will recursively exclude all artifacts
    provided by the given dep.
    """
    if isinstance(dep, JarDependency):
      yield Exclude(dep.org, dep.name)
    elif isinstance(dep, ExportableJvmLibrary):
      if not dep.provides:
        raise TargetDefinitionException(self,
            'Targets passed to `overrides` must represent published artifacts. %s does not.' % dep)
      yield Exclude(dep.provides.org, dep.provides.name)
    elif isinstance(dep, JarLibrary):
      for d in dep._pre_override_dependencies:
        for exclude in self._excludes(d):
          yield exclude

  def _jar_dependencies(self, dep):
    """
    A generator for JarDependencies transitively included by the given dep.
    """
    if isinstance(dep, JarDependency):
      yield dep
    elif isinstance(dep, JarLibrary):
      for direct_dep in dep._pre_override_dependencies:
        for dep in self._jar_dependencies(direct_dep):
          yield dep
    elif isinstance(dep, Pants):
      for d in self._jar_dependencies(dep.get()):
        yield d



import hashlib
import os

from collections import defaultdict
from contextlib import contextmanager

from twitter.common.collections import OrderedSet
from twitter.common.dirutil import Lock

from twitter.pants import SourceRoot
from twitter.pants.base import ParseContext
from twitter.pants.base.target import Target
from twitter.pants.targets import Pants
from twitter.pants.goal.products import Products

class Context(object):
  """Contains the context for a single run of pants.

  Goal implementations can access configuration data from pants.ini and any flags they have exposed
  here as well as information about the targets involved in the run.

  Advanced uses of the context include adding new targets to it for upstream or downstream goals to
  operate on and mapping of products a goal creates to the targets the products are associated with.
  """

  class Log(object):
    def debug(self, msg): pass
    def info(self, msg): pass
    def warn(self, msg): pass
    def error(self, msg): pass

  def __init__(self, config, options, target_roots, lock=None, log=None, target_base=None, buildtime_stats=None):
    self._config = config
    self._options = options
    self._lock = lock or Lock.unlocked()
    self._log = log or Context.Log()
    self._target_base = target_base or Target
    self._state = {}
    self._products = Products()
    self.buildtime_stats = buildtime_stats

    self.replace_targets(target_roots)

  @property
  def config(self):
    """Returns a Config object containing the configuration data found in pants.ini."""
    return self._config

  @property
  def options(self):
    """Returns the command line options parsed at startup."""
    return self._options

  @property
  def lock(self):
    """Returns the global pants run lock so a goal can release it if needed."""
    return self._lock

  @property
  def log(self):
    """Returns the preferred logger for goals to use."""
    return self._log

  @property
  def products(self):
    """Returns the Products manager for the current run."""
    return self._products

  @property
  def target_roots(self):
    """Returns the targets specified on the command line.

    This set is strictly a subset of all targets in play for the run as returned by self.targets().
    Note that for a command line invocation that uses wildcard selectors : or ::, the targets
    globbed by the wildcards are considered to be target roots.
    """
    return self._target_roots

  def identify(self, targets):
    id = hashlib.md5()
    for target in targets:
      id.update(target.id)
    return id.hexdigest()

  def __str__(self):
    return 'Context(id:%s, state:%s, targets:%s)' % (self.id, self.state, self.targets())

  def replace_targets(self, target_roots):
    """Replaces all targets in the context with the given roots and their transitive
    dependencies.
    """
    self._target_roots = list(target_roots)

    self._targets = OrderedSet()
    for target in self._target_roots:
      self.add_target(target)
    self.id = self.identify(self._targets)

  def add_target(self, target):
    """Adds a target and its transitive dependencies to the run context.

    The target is not added to the target roots.
    """
    def add_targets(tgt):
      self._targets.update(tgt for tgt in tgt.resolve() if isinstance(tgt, self._target_base))
    target.walk(add_targets)

  def add_new_target(self, target_base, target_type, *args, **kwargs):
    """Creates a new target, adds it to the context and returns it.

    This method ensures the target resolves files against the given target_base, creating the
    directory if needed and registering a source root.
    """
    target = self._create_new_target(target_base, target_type, *args, **kwargs)
    self.add_target(target)
    return target

  def _create_new_target(self, target_base, target_type, *args, **kwargs):
    if not os.path.exists(target_base):
      os.makedirs(target_base)
    SourceRoot.register(target_base, target_type)
    with ParseContext.temp(target_base):
      return target_type(*args, **kwargs)

  def remove_target(self, target):
    """Removes the given Target object from the context completely if present."""
    if target in self.target_roots:
      self.target_roots.remove(target)
    self._targets.discard(target)

  def targets(self, predicate=None):
    """Selects targets in-play in this run from the target roots and their transitive dependencies.

    If specified, the predicate will be used to narrow the scope of targets returned.
    """
    return filter(predicate, self._targets)

  def dependants(self, on_predicate=None, from_predicate=None):
    """Returns  a map from targets that satisfy the from_predicate to targets they depend on that
      satisfy the on_predicate.
    """
    core = set(self.targets(on_predicate))
    dependees = defaultdict(set)
    for target in self.targets(from_predicate):
      if hasattr(target, 'dependencies'):
        for dependency in target.dependencies:
          if dependency in core:
            dependees[target].add(dependency)
    return dependees

  def resolve(self, spec):
    """Returns an iterator over the target(s) the given address points to."""
    with ParseContext.temp():
      return Pants(spec).resolve()

  @contextmanager
  def state(self, key, default=None):
    value = self._state.get(key, default)
    yield value
    self._state[key] = value



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from collections import defaultdict

import os
import shlex

from twitter.common.dirutil import safe_open, safe_mkdir

from twitter.pants import has_sources, is_apt
from twitter.pants.tasks import TaskError
from twitter.pants.tasks.jvm_compiler_dependencies import Dependencies
from twitter.pants.tasks.nailgun_task import NailgunTask


# Well known metadata file to auto-register annotation processors with a java 1.6+ compiler
_PROCESSOR_INFO_FILE = 'META-INF/services/javax.annotation.processing.Processor'


_JMAKE_MAIN = 'com.sun.tools.jmake.Main'


# From http://kenai.com/projects/jmake/sources/mercurial/content/src/com/sun/tools/jmake/Main.java?rev=26
# Main.mainExternal docs.
_JMAKE_ERROR_CODES = {
   -1: 'invalid command line option detected',
   -2: 'error reading command file',
   -3: 'project database corrupted',
   -4: 'error initializing or calling the compiler',
   -5: 'compilation error',
   -6: 'error parsing a class file',
   -7: 'file not found',
   -8: 'I/O exception',
   -9: 'internal jmake exception',
  -10: 'deduced and actual class name mismatch',
  -11: 'invalid source file extension',
  -12: 'a class in a JAR is found dependent on a class with the .java source',
  -13: 'more than one entry for the same class is found in the project',
  -20: 'internal Java error (caused by java.lang.InternalError)',
  -30: 'internal Java error (caused by java.lang.RuntimeException).'
}
# When executed via a subprocess return codes will be treated as unsigned
_JMAKE_ERROR_CODES.update((256+code, msg) for code, msg in _JMAKE_ERROR_CODES.items())


def _is_java(target):
  return has_sources(target, '.java')


class JavaCompile(NailgunTask):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    NailgunTask.setup_parser(option_group, args, mkflag)

    option_group.add_option(mkflag("warnings"), mkflag("warnings", negate=True),
                            dest="java_compile_warnings", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Compile java code with all configured warnings "
                                 "enabled.")

    option_group.add_option(mkflag("args"), dest="java_compile_args", action="append",
                            help="Pass these extra args to javac.")

  def __init__(self, context):
    super(JavaCompile, self).__init__(context)

    workdir = context.config.get('java-compile', 'workdir')
    self._classes_dir = os.path.join(workdir, 'classes')
    self._resources_dir = os.path.join(workdir, 'resources')
    self._dependencies_file = os.path.join(workdir, 'dependencies')

    self._jmake_profile = context.config.get('java-compile', 'jmake-profile')
    self._compiler_profile = context.config.get('java-compile', 'compiler-profile')

    self._opts = context.config.getlist('java-compile', 'args')
    self._jvm_args = context.config.getlist('java-compile', 'jvm_args')

    self._javac_opts = []
    if context.options.java_compile_args:
      for arg in context.options.java_compile_args:
        self._javac_opts.extend(shlex.split(arg))
    else:
      self._javac_opts.extend(context.config.getlist('java-compile', 'javac_args', default=[]))

    if context.options.java_compile_warnings:
      self._opts.extend(context.config.getlist('java-compile', 'warning_args'))
    else:
      self._opts.extend(context.config.getlist('java-compile', 'no_warning_args'))

    self._confs = context.config.getlist('java-compile', 'confs')

  def execute(self, targets):
    java_targets = filter(_is_java, targets)
    if java_targets:
      jmake_classpath = self.profile_classpath(self._jmake_profile)
      compiler_classpath = self.profile_classpath(self._compiler_profile)

      with self.context.state('classpath', []) as cp:
        for conf in self._confs:
          cp.insert(0, (conf, self._classes_dir))

        with self.changed(java_targets, invalidate_dependants=True) as changed:
          sources_by_target, processors, fingerprint = self.calculate_sources(changed)
          if sources_by_target:
            sources = reduce(lambda all_, sources: all_.union(sources), sources_by_target.values())
            if not sources:
              self.context.log.warn('Skipping java compile for targets with no sources:\n  %s' %
                                    '\n  '.join(str(t) for t in sources_by_target.keys()))
            else:
              classpath = [jar for conf, jar in cp if conf in self._confs]
              result = self.compile(jmake_classpath, compiler_classpath, classpath, sources,
                                    fingerprint)
              if result != 0:
                task_error_msg = 'java %s ... exited non-zero (%i)' % (_JMAKE_MAIN, result)
                error_msg = _JMAKE_ERROR_CODES.get(result, None)

                if error_msg:
                  task_error_msg += " '%s'" % error_msg
                raise TaskError(task_error_msg)

            if processors:
              # Produce a monolithic apt processor service info file for further compilation rounds
              # and the unit test classpath.
              processor_info_file = os.path.join(self._classes_dir, _PROCESSOR_INFO_FILE)
              if os.path.exists(processor_info_file):
                with safe_open(processor_info_file, 'r') as f:
                  for processor in f:
                    processors.add(processor.strip())
              self.write_processor_info(processor_info_file, processors)

      if self.context.products.isrequired('classes'):
        genmap = self.context.products.get('classes')

        # Map generated classes to the owning targets and sources.
        dependencies = Dependencies(self._classes_dir, self._dependencies_file)
        for target, classes_by_source in dependencies.findclasses(targets).items():
          for source, classes in classes_by_source.items():
            genmap.add(source, self._classes_dir, classes)
            genmap.add(target, self._classes_dir, classes)

        # 'Map' (rewrite) annotation processor service info files to the owning targets.
        for target in targets:
          if is_apt(target) and target.processors:
            basedir = os.path.join(self._resources_dir, target.id)
            processor_info_file = os.path.join(basedir, _PROCESSOR_INFO_FILE)
            self.write_processor_info(processor_info_file, target.processors)
            genmap.add(target, basedir, [_PROCESSOR_INFO_FILE])

  def calculate_sources(self, targets):
    sources = defaultdict(set)
    processors = set()

    def collect_sources(target):
      src = (os.path.join(target.target_base, source)
             for source in target.sources if source.endswith('.java'))
      if src:
        sources[target].update(src)
        if is_apt(target) and target.processors:
          processors.update(target.processors)

    for target in targets:
      collect_sources(target)
    return sources, processors, self.context.identify(targets)

  def compile(self, jmake_classpath, compiler_classpath, classpath, sources, fingerprint):
    safe_mkdir(self._classes_dir)

    args = [
      '-classpath', ':'.join(classpath),
      '-d', self._classes_dir,
      '-pdb', os.path.join(self._classes_dir, '%s.dependencies.pdb' % fingerprint),
    ]

    args.extend([
      '-jcpath', ':'.join(compiler_classpath),
      '-jcmainclass', 'com.twitter.common.tools.Compiler',
      '-C-Tdependencyfile', '-C%s' % self._dependencies_file,
    ])
    args.extend(map(lambda arg: '-C%s' % arg, self._javac_opts))

    args.extend(self._opts)
    args.extend(sources)
    return self.runjava(jmake_classpath, _JMAKE_MAIN, args=args, jvm_args=self._jvm_args)

  def write_processor_info(self, processor_info_file, processors):
    with safe_open(processor_info_file, 'w') as f:
      for processor in processors:
        f.write('%s\n' % processor)



__author__ = 'Anand Madhavan'

# TODO(Anand) Remove this from pants proper when a code adjoinment mechanism exists
# or ok if/when thriftstore is open sourced as well

import os
import subprocess

from twitter.common.collections import OrderedSet
from twitter.common.dirutil import safe_mkdir

from twitter.pants import is_jvm
from twitter.pants.targets import JavaLibrary, JavaThriftstoreDMLLibrary

from .code_gen import CodeGen

from . import TaskError


class ThriftstoreDMLGen(CodeGen):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("outdir"), dest="thriftstore_gen_create_outdir",
                            help="Emit thriftstore generated code in to this directory.")

  def __init__(self, context):
    CodeGen.__init__(self, context)
    self.thriftstore_codegen = context.config.get('thriftstore-dml-gen', 'thriftstore-codegen')

    self.output_dir = (context.options.thriftstore_gen_create_outdir or
                       context.config.get('thriftstore-dml-gen', 'workdir'))

    self.verbose = context.config.getbool('thriftstore-dml-gen', 'verbose')

    def create_javadeps():
      gen_info = context.config.getlist('thriftstore-dml-gen', 'javadeps')
      deps = OrderedSet()
      for dep in gen_info:
        deps.update(context.resolve(dep))
      return deps

    def is_thriftstore_dml_instance(target):
      return isinstance(target, JavaThriftstoreDMLLibrary)

    # Resolved java library targets go in javadeps
    self.javadeps = create_javadeps()

    self.gen_thriftstore_java_dir = os.path.join(self.output_dir, 'gen-thriftstore-java')

    def insert_java_dml_targets():
      self.gen_dml_jls = {}
      # Create a synthetic java library for each dml target
      for thrift_dml_lib in context.targets(is_thriftstore_dml_instance):
        # Add one generated JavaLibrary target (whose sources we will fill in later on)
        java_dml_lib = self.context.add_new_target(self.gen_thriftstore_java_dir,
                                                   JavaLibrary,
                                                   name=thrift_dml_lib.id,
                                                   sources=[],
                                                   dependencies=self.javadeps)
        java_dml_lib.is_codegen = True
        java_dml_lib.update_dependencies([thrift_dml_lib])
        self.gen_dml_jls[thrift_dml_lib] = java_dml_lib

      for dependee, dmls in context.dependants(is_thriftstore_dml_instance).items():
        for jl in map(lambda dml: self.gen_dml_jls[dml], dmls):
          self.updatedependencies(dependee, jl)

    insert_java_dml_targets()

  def invalidate_for(self):
    return set('java')

  def is_gentarget(self, target):
    return isinstance(target, JavaThriftstoreDMLLibrary)

  def is_forced(self, lang):
    return True

  def genlangs(self):
    return dict(java=is_jvm)

  def genlang(self, lang, targets):
    bases, sources = self._calculate_sources(targets)

    safe_mkdir(self.gen_thriftstore_java_dir)

    args = [
      self.thriftstore_codegen,
      'dml',
      '-o', self.gen_thriftstore_java_dir
    ]

    if self.verbose:
      args.append('-verbose')
    args.extend(sources)
    self.context.log.debug('Executing: %s' % ' '.join(args))
    result = subprocess.call(args)
    if result != 0:
      raise TaskError('%s ... exited non-zero (%i)' % (self.thriftstore_codegen, result))

  def _calculate_sources(self, thrift_targets):
    bases = set()
    sources = set()
    def collect_sources(target):
      if self.is_gentarget(target):
        bases.add(target.target_base)
        sources.update(os.path.join(target.target_base, source) for source in target.sources)
    for target in thrift_targets:
      target.walk(collect_sources)
    return bases, sources

  def createtarget(self, lang, gentarget, dependees):
    if lang == 'java':
      return self._create_java_target(gentarget)
    else:
      raise TaskError('Unrecognized thrift gen lang: %s' % lang)

  def _calculate_genfiles(self, sources):
    args = [
      self.thriftstore_codegen,
      'parse'
    ]
    args.extend(sources)
    self.context.log.debug('Executing: %s' % ' '.join(args))
    p = subprocess.Popen(args,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
    output, error = p.communicate()
    result = p.wait()
    if result != 0:
      raise TaskError('%s ... exited non-zero (%i)' % (self.thriftstore_codegen, result))
    thriftstore_classes = filter(lambda s: s.strip() != '', output.split('\n'))
    self.context.log.debug('Generated files: \n\t\t%s' % '\n\t\t'.join(thriftstore_classes))
    return thriftstore_classes

  def _create_java_target(self, target):
    source_files = [os.path.join(target.target_base, source) for source in target.sources]
    self.gen_dml_jls[target].sources = self._calculate_genfiles(source_files)
    return self.gen_dml_jls[target]



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from .distribution import Distribution
from .executor import Executor, SubprocessExecutor
from .nailgun_client import NailgunClient
from .nailgun_executor import NailgunExecutor


def runjava(classpath, main, args=None, jvm_args=None, **subprocess_kwargs):
  """Spawns a java process with the supplied configuration and returns its exit code.

  Any extra kwargs are passed through to ``subprocess.Popen`` to control the spawned process
  environment further,

  :param list classpath: the classpath for the java program
  :param string main: the fully qualified class name of the java program's entry point
  :param list args: an optional sequence of args to pass to the java program
  :param list jvm_args: an optional sequence of args for the underlying jvm
  """
  java = SubprocessExecutor()
  return java.spawn(classpath, main, args=args, jvm_args=jvm_args, **subprocess_kwargs).wait()


__all__ = (
    'runjava',
    'Distribution',
    'Executor',
    'NailgunClient',
    'NailgunExecutor',
    'SubprocessExecutor',
)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.common.collections import maybe_list

from twitter.pants.base import manual, Target
from twitter.pants.base.target import TargetDefinitionException
from twitter.pants.targets.exportable_jvm_library import ExportableJvmLibrary
from twitter.pants.targets.resources import WithLegacyResources

from . import JavaLibrary


@manual.builddict(tags=["scala"])
class ScalaLibrary(ExportableJvmLibrary, WithLegacyResources):
  """A collection of Scala code.

  Normally has conceptually-related sources; invoking the ``compile`` goal
  on this target compiles scala and generates classes. Invoking the ``bundle``
  goal on this target creates a ``.jar``; but that's an unusual thing to do.
  Instead, a ``jvm_binary`` might depend on this library; that binary is a
  more sensible thing to bundle.
  """

  def __init__(self, name,
               sources = None,
               java_sources = None,
               provides = None,
               dependencies = None,
               excludes = None,
               resources = None,
               deployjar = False,
               buildflags = None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param sources: A list of filenames representing the source code
      this library is compiled from.
    :type sources: list of strings
    :param java_sources:
      :class:`twitter.pants.targets.java_library.JavaLibrary` or list of
      JavaLibrary targets this library has a circular dependency on.
      Prefer using dependencies to express non-circular dependencies.
    :param Artifact provides:
      The :class:`twitter.pants.targets.artifact.Artifact`
      to publish that represents this target outside the repo.
    :param dependencies: List of :class:`twitter.pants.base.target.Target` instances
      this target depends on.
    :type dependencies: list of targets
    :param excludes: List of :class:`twitter.pants.targets.exclude.Exclude` instances
      to filter this target's transitive dependencies against.
    :param resources: An optional list of paths (DEPRECATED) or ``resources``
      targets containing resources that belong on this library's classpath.
    :param deployjar: Unused, and will be removed in a future release.
    :param buildflags: Unused, and will be removed in a future release.
    """

    ExportableJvmLibrary.__init__(self,
                                  name,
                                  sources,
                                  provides,
                                  dependencies,
                                  excludes,
                                  buildflags)

    WithLegacyResources.__init__(self, name, sources=sources, resources=resources)

    if (sources is None) and (resources is None):
      raise TargetDefinitionException(self, 'Must specify sources and/or resources.')

    self.deployjar = deployjar

    # Defer resolves until done parsing the current BUILD file, certain source_root arrangements
    # might allow java and scala sources to co-mingle and so have targets in the same BUILD.
    self._post_construct(self._link_java_cycles, java_sources)

  def _link_java_cycles(self, java_sources):
    if java_sources:
      self.java_sources = list(Target.resolve_all(maybe_list(java_sources, Target), JavaLibrary))
    else:
      self.java_sources = []

    # We have circular java/scala dep, add an inbound dependency edge from java to scala in this
    # case to force scala compilation to precede java - since scalac supports generating java stubs
    # for these cycles and javac does not this is both necessary and always correct.
    for java_target in self.java_sources:
      java_target.update_dependencies([self])



# ==================================================================================================
# Copyright 2014 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.pants.targets.exclude import Exclude
from twitter.pants.targets.jvm_binary import JvmBinary

# The hadoop cluster provides the appropriate versions of these jars.
ON_HADOOP_CLUSTER_EXCLUDES = [
  Exclude('commons-httpclient','commons-httpclient'),
  Exclude('org.apache', 'hadoop-core'),
  Exclude('org.apache', 'hadoop-lzo'),
  Exclude('com.hadoop', 'hadoop-lzo'),
  Exclude('org.apache.hadoop')
]

class TwitterHadoopBinary(JvmBinary):
  """A binary that is suitable for running on the Hadoop cluster.

  Invoking the ``binary`` or ``bundle`` goal on one of these targets creates a binary jar that
  excludes any dependencies already provided by the hadoop cluster.
  """
  def __init__(
      self,
      name,
      main,
      basename=None,
      source=None,
      resources=None,
      dependencies=None,
      excludes=[],
      deploy_excludes=[]):

    JvmBinary.__init__(
      self,
      name=name,
      main=main,
      basename=basename,
      source=source,
      resources=resources,
      dependencies=dependencies,
      excludes=excludes,
      deploy_excludes=(deploy_excludes or []) + ON_HADOOP_CLUSTER_EXCLUDES,
      configurations=None
    )



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.pants.base import manual, Target

from .external_dependency import ExternalDependency

from pkg_resources import Requirement


@manual.builddict(tags=["python"])
class PythonRequirement(Target, ExternalDependency):
  """Pants wrapper around pkg_resources.Requirement"""

  def __init__(self,
               requirement,
               name=None,
               repository=None,
               version_filter=None,
               use_2to3=False,
               compatibility=None):
    # TODO(wickman) Allow PythonRequirements to be specified using pip-style vcs or url identifiers,
    # e.g. git+https or just http://...
    self._requirement = Requirement.parse(requirement)
    self._repository = repository
    self._name = name or self._requirement.project_name
    self._use_2to3 = use_2to3
    self._version_filter = version_filter or (lambda py, pl: True)
    # TODO(wickman) Unify this with PythonTarget .compatibility
    self.compatibility = compatibility or ['']
    Target.__init__(self, self._name)

  def should_build(self, python, platform):
    return self._version_filter(python, platform)

  @property
  def use_2to3(self):
    return self._use_2to3

  @property
  def repository(self):
    return self._repository

  # duck-typing Requirement interface for Resolver, since Requirement cannot be
  # subclassed (curses!)
  @property
  def key(self):
    return self._requirement.key

  @property
  def extras(self):
    return self._requirement.extras

  @property
  def specs(self):
    return self._requirement.specs

  @property
  def project_name(self):
    return self._requirement.project_name

  @property
  def requirement(self):
    return self._requirement

  def __contains__(self, item):
    return item in self._requirement

  def cache_key(self):
    return str(self._requirement)

  def __repr__(self):
    return 'PythonRequirement(%s)' % self._requirement



import os

from collections import defaultdict

from twitter.common.dirutil import safe_mkdir
from twitter.pants import get_buildroot, is_java
from twitter.pants.fs.archive import TGZ
from twitter.pants.targets.oink_query import OinkQuery
from twitter.pants.tasks import Task
from twitter.pants.tasks.jvm_binary_task import  JvmBinaryTask

class OinkBundleCreate(JvmBinaryTask):
  """Produces a deployable tar.gz bundle for an Oink Query.

  This is invoked as a bundle goal against OinkQuery Target.

  Example usage::

    ./pants goal bundle /path/to/oink_query_target
  """

  @staticmethod
  def is_oink_query(target):
    return isinstance(target, OinkQuery)

  def __init__(self, context):
    JvmBinaryTask.__init__(self, context)
    self.outdir = context.config.getdefault('outdir')
    self.context.products.require('jars', predicate=is_java)
    self.context.products.require('jar_dependencies', predicate=self.is_oink_query)

  def execute(self, targets):
    all_jars = []
    def _visitor(target):
      jar_products = self.context.products.get('jars').get(target)
      if jar_products is not None:
        all_jars.append(jar_products)

    for oink_query in filter(self.is_oink_query, targets):
      oink_query.walk(_visitor, is_java)

      dependencies_with_excludes = defaultdict(list)
      # list_jar_dependencies below expects a JvmBinary but we can safely pass an OinkQuery
      # since it has the deploy_excludes property and walk method list_jar_dependencies expects
      for basedir, externaljar in self.list_jar_dependencies(oink_query):
        dependencies_with_excludes[basedir].append(externaljar)
      all_jars.append(dependencies_with_excludes)

      self.context.log.debug("all jar dependencies: %s" % all_jars)

      flattened_jar_paths = []
      for product_mapping in all_jars:
        for basedir, jars in product_mapping.iteritems():
          flattened_jar_paths.extend(os.path.join(basedir, jar) for jar in jars)

      bundledir = os.path.join(self.outdir, '%s-bundle' % oink_query.name)
      libs_bundle_dir = os.path.join(bundledir, 'libs')

      safe_mkdir(bundledir, clean=True)

      safe_mkdir(libs_bundle_dir)
      for path in flattened_jar_paths:
        os.symlink(path, os.path.join(libs_bundle_dir, os.path.basename(path)))

      os.symlink('%s/src/main/pig' % get_buildroot(), os.path.join(bundledir, 'pig'))
      os.symlink('%s/src/main/ruby/gems' % get_buildroot(), os.path.join(bundledir, 'gems'))
      os.symlink('%s/src/main/ruby/oink' % get_buildroot(), os.path.join(bundledir, 'oink'))
      os.symlink('%s/src/scripts/oink_bundle/oink_bundle_runner.sh' % get_buildroot(),
                 os.path.join(bundledir, 'oink_bundle_runner.sh'))

      for source in oink_query.sources:
        os.symlink(os.path.join(get_buildroot(), oink_query.target_base, source),
                   os.path.join(bundledir, source))

      tarpath = TGZ.create(bundledir, self.outdir, oink_query.name)
      self.context.log.info('created %s' % os.path.relpath(tarpath, get_buildroot()))



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import getpass
import os
import subprocess

from twitter.common.config import Properties
from twitter.common.dirutil import safe_mkdir
from twitter.pants.targets import RubyThriftLibrary
from twitter.pants.tasks.scm_publish import ScmPublish, Semver
from twitter.pants import get_scm
from . import Task, TaskError


def is_ruby_library(target):
  return isinstance(target, RubyThriftLibrary)


class PushDb(object):
  @staticmethod
  def load(path):
    """Loads a pushdb maintained in a properties file at the given path."""
    with open(path, 'r') as props:
      properties = Properties.load(props)
      return PushDb(properties)

  def __init__(self, props):
    self._props = props

  def next_version(self, target):
    """
      Given an internal target, return a JarDependency with the last published revision filled in.
    """
    return Semver.parse(self._props.get(target.provides.name) or "0.0.0").bump().version()

  def set_version(self, target, version):
    self._props[target.provides.name] = version

  def dump(self, path):
    """Saves the pushdb as a properties file to the given path."""
    with open(path, 'w') as props:
      Properties.dump(self._props, props)


class ThriftGemerator(ScmPublish, Task):
  """Create ruby gems from thrift IDL files.

  ThriftGemerator can be configured in ``pants.ini`` as follows.

  ``gem-create`` section:

  * ``restrict_push_branches`` - Optional list of branches to restrict publishing to.
  * ``gems`` - Optional path of directory to create gems in.
  """

  _CONFIG_SECTION = 'gem-create'

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("outdir"), dest="thrift_gemerator_outdir",
                            help="Emit gems in to this directory.")

    option_group.add_option(mkflag("override"), dest="thrift_gemerator_override",
                            help="Version number override.")

    option_group.add_option(mkflag("dryrun"), mkflag("dryrun", negate=True),
                            dest="gem_publish_dryrun", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Runs through gemerator without actually pushing "
                                 "artifacts, editing publish dbs or otherwise writing data")

    option_group.add_option(mkflag("commit", negate=True),
                            dest="gem_publish_commit", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="Turns off commits of the push db for local testing.")

  def __init__(self, context, scm=None):
    Task.__init__(self, context)
    ScmPublish.__init__(self, scm or get_scm(),
                        self.context.config.getlist(
                          ThriftGemerator._CONFIG_SECTION, 'restrict_push_branches'))
    options = context.options
    self._output_dir = (options.thrift_gemerator_outdir or
                        self.get_workdir(section=ThriftGemerator._CONFIG_SECTION, workdir='gems'))
    self._override = options.thrift_gemerator_override
    self._dryrun = options.gem_publish_dryrun
    self._commit = options.gem_publish_commit

  def execute(self, targets):
    self.check_clean_master(commit=(not self._dryrun and self._commit))

    safe_mkdir(self._output_dir)

    self.gem(filter(is_ruby_library, self.context.target_roots))

  def gem(self, targets):
    for target in targets:
      thrift_files_with_root = []
      for source in target.derived_from.sources:
        thrift_files_with_root.append(os.path.join(target.derived_from.target_base , source))
      thrift_files = " ".join(thrift_files_with_root)
      gem_name = target.provides.name
      push_db = PushDb.load(target.provides.repo.push_db)
      gem_version = self._override or push_db.next_version(target)

      bases = set()
      for target in targets:
        bases.add(target.target_base)


      args = dict(
        include='-I "%s"' % '" -I "'.join(bases),
        name=gem_name,
        outdir=self._output_dir,
        version=gem_version,
        thrift_files=thrift_files,
        user=getpass.getuser(),
        publish=' --publish %s' % target.provides.repo.url if not self._dryrun else ''
      )

      cmd = 'thrift_gemerator --no-interactive --compiler-args="%(include)s" --gem=%(name)s' \
            ' --output=%(outdir)s --gem-version %(version)s%(publish)s' \
            ' --no-git-repo %(thrift_files)s' % args
      log = self.context.log
      log.debug("Executing: %s" % cmd)
      result = subprocess.call(cmd, shell=True)
      if result != 0:
        raise TaskError('ThriftGemerator exited non-zero %i' % result)

      if self._dryrun:
        log.debug("Skipping gem publish for %s, but gem can be found at %s" % (gem_name, self._output_dir))
      else:
        push_db.set_version(target, gem_version)
        push_db.dump(target.provides.repo.push_db)

      if self._dryrun or not self._commit:
        log.debug("Skipping commit for %s" % gem_name)
      else:
        self.commit_push("gem %s" % gem_name)
        self.scm.refresh()
        self.scm.tag('gem-%(name)s-%(version)s' % args,
                     message='Publish of gem %(name)s-%(version)s initiated by %(user)s' % args)



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from collections import Sequence

from twitter.common.dirutil.fileset import Fileset
from twitter.common.lang import Compatibility
from twitter.pants import is_concrete
from twitter.pants.base import ParseContext, Target
from twitter.pants.targets.internal import InternalTarget
from twitter.pants.targets.with_sources import TargetWithSources


class Resources(InternalTarget, TargetWithSources):
  """Describes a set of resource files to be embedded in a library or binary.

  If your target compiles to the JVM (e.g., ``java_library``,
  ``scala_library``, ``junit_tests``), you might have files
  that you need to access as resources. Each of these targets has an optional
  argument called ``resources`` that expects a list of target addresses that
  resolve to targets whose type is resource.

  In the ``jar`` goal, the resource files are placed in the resulting `.jar`.
  """

  def __init__(self, name, sources):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param sources: A list of filenames representing the resources
      this library provides.
    """
    InternalTarget.__init__(self, name, dependencies=[])
    TargetWithSources.__init__(self, name)

    self.sources = self._resolve_paths(self.target_base, sources)


class WithLegacyResources(TargetWithSources):
  """Collects resources whether they are specified using globs against an assumed parallel
  'resources' directory or they are Resources targets.

  Resource handling is currently in transition, and two forms are available.

  New style:

  ::

    # `resources' is a first-class target type.
    resources(name='mybird',
      sources=['list', 'of', 'resources']
    )

    # Dependees depend on one or more `resources' targets.
    resources=[pants('src/resources/com/twitter/mybird')]

  Old style:

  ::

    # Resources can be a fileset.
    resources=globs('*.txt')

    # Resources can be a list of filenames.
    resources=['list', 'of', 'files']

  Please note mixing old/new styles is not supported.

  """

  def __init__(self, name, sources=None, resources=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param sources: A list of filenames representing the source code
      this library is compiled from.
    :param resources: One or more :class:`twitter.pants.targets.resources.Resources`
      xor a list of filenames representing the resources this library provides.
    """
    TargetWithSources.__init__(self, name, sources=sources)

    if resources is not None:
      def is_resources(item):
        return (isinstance(item, Target)
                and all(map(lambda tgt: isinstance(tgt, Resources),
                            filter(lambda tgt: is_concrete(tgt), item.resolve()))))

      resources_seq = resources if isinstance(resources, Sequence) else [resources]
      if all(map(is_resources, resources_seq)):
        self.resources = list(self.resolve_all(resources_seq, Resources))
      elif (all(map(lambda resource: isinstance(resource, Fileset)
            or isinstance(resource, Compatibility.string), resources_seq))):
        # Handle parallel resource dir globs.
        # For example, for a java_library target base of src/main/java:
        #   src/main/java/com/twitter/base/BUILD
        # We get:
        #   sibling_resources_base = src/main/resources
        #   base_relpath = com/twitter/base
        #   resources_dir = src/main/resources/com/twitter/base
        #
        # TODO(John Sirois): migrate projects to Resources and remove support for old style assumed
        # parallel resources dirs
        sibling_resources_base = os.path.join(os.path.dirname(self.target_base), 'resources')
        base_relpath = os.path.relpath(self.address.buildfile.relpath, self.target_base)
        resources_dir = os.path.join(sibling_resources_base, base_relpath)
        with ParseContext.temp(basedir=resources_dir):
          self.resources = [Resources(name, resources)]
      else:
        raise ValueError('Target %s resources are invalid: %s' % (self.address, resources))

      # Add the resources to dependencies.
      if isinstance(self, InternalTarget):
        self.update_dependencies(self.resources)


# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from abc import abstractproperty

from twitter.common.lang import AbstractClass

from .jar_dependency import JarDependency


class Jarable(AbstractClass):
  """A mixin that identifies a target as one that can provide a jar."""

  @abstractproperty
  def identifier(self):
    """Subclasses should return a stable unique identifier for the jarable target."""

  @property
  def provides(self):
    """Returns an optional :class:`twitter.pants.targets.Artifact` if this target is exportable.

    Subclasses should override to provide an artifact descriptor when one applies, by default None
    is supplied.
    """
    return None

  def get_artifact_info(self):
    """Returns a triple composed of a :class:`twitter.pants.targets.jar_dependency.JarDependency`
    describing the jar for this target, this target's artifact identifier and a bool indicating if
    this target is exportable.
    """
    exported = bool(self.provides)

    org = self.provides.org if exported else 'internal'
    module = self.provides.name if exported else self.identifier

    id_ = "%s-%s" % (self.provides.org, self.provides.name) if exported else self.identifier

    # TODO(John Sirois): This should return something less than a JarDependency encapsulating just
    # the org and name.  Perhaps a JarFamily?
    return JarDependency(org=org, name=module, rev=None), id_, exported



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import hashlib
import os
import re
import tempfile

from collections import defaultdict, namedtuple

from twitter.common.collections import OrderedSet
from twitter.common.dirutil import safe_mkdir, safe_open

from twitter.pants import get_buildroot
from twitter.pants.targets import InternalTarget, JavaLibrary, JavaThriftLibrary, ScalaLibrary
from twitter.pants.tasks import TaskError
from twitter.pants.tasks.nailgun_task import NailgunTask
from twitter.pants.thrift_util import (
  calculate_compile_sources,
  calculate_compile_sources_HACK_FOR_SCROOGE_LEGACY)

CompilerConfig = namedtuple('CompilerConfig', ['name', 'config_section', 'profile',
                                               'main', 'calc_srcs', 'langs'])


class Compiler(namedtuple('CompilerConfigWithContext', ('context',) + CompilerConfig._fields)):
  @classmethod
  def fromConfig(cls, context, config):
    return cls(context, **config._asdict())

  @property
  def jvm_args(self):
    args = self.context.config.getlist(self.config_section, 'jvm_args', default=[])
    args.append('-Dfile.encoding=UTF-8')
    return args

  @property
  def outdir(self):
    pants_workdir_fallback = os.path.join(get_buildroot(), '.pants.d')
    workdir_fallback = os.path.join(self.context.config.getdefault('pants_workdir',
                                                                   default=pants_workdir_fallback),
                                    self.name)
    outdir = (self.context.options.scrooge_gen_create_outdir
              or self.context.config.get(self.config_section, 'workdir', default=workdir_fallback))
    return os.path.relpath(outdir)

  @property
  def verbose(self):
    if self.context.options.scrooge_gen_quiet is not None:
      return not self.context.options.scrooge_gen_quiet
    else:
      return self.context.config.getbool(self.config_section, 'verbose', default=False)

  @property
  def strict(self):
    return self.context.config.getbool(self.config_section, 'strict', default=False)


_COMPILERS = [
    CompilerConfig(name='scrooge',
                   config_section='scrooge-gen',
                   profile='scrooge-gen',
                   main='com.twitter.scrooge.Main',
                   calc_srcs=calculate_compile_sources,
                   langs=frozenset(['scala', 'java'])),
    CompilerConfig(name='scrooge-legacy',
                   config_section='scrooge-legacy-gen',
                   profile='scrooge-legacy-gen',
                   main='com.twitter.scrooge.Main',
                   calc_srcs=calculate_compile_sources_HACK_FOR_SCROOGE_LEGACY,
                   langs=frozenset(['scala']))
]

_CONFIG_FOR_COMPILER = dict((compiler.name, compiler) for compiler in _COMPILERS)

_TARGET_TYPE_FOR_LANG = dict(scala=ScalaLibrary, java=JavaLibrary)


class ScroogeGen(NailgunTask):
  GenInfo = namedtuple('GenInfo', ['gen', 'deps'])

  class PartialCmd(namedtuple('PC', ['compiler', 'language', 'rpc_style', 'namespace_map'])):
    @property
    def outdir(self):
      namespace_sig = None
      if self.namespace_map:
        sha = hashlib.sha1()
        for ns_from, ns_to in sorted(self.namespace_map):
          sha.update(ns_from)
          sha.update(ns_to)
        namespace_sig = sha.hexdigest()
      output_style = '-'.join(filter(None, (self.language, self.rpc_style, namespace_sig)))
      return os.path.join(self.compiler.outdir, output_style)

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("outdir"), dest="scrooge_gen_create_outdir",
                            help="Emit generated code in to this directory.")
    option_group.add_option(mkflag("quiet"), dest="scrooge_gen_quiet",
                            action="callback", callback=mkflag.set_bool, default=None,
                            help="[%default] Suppress output, overrides verbose flag in pants.ini.")

  def __init__(self, context):
    super(ScroogeGen, self).__init__(context)
    self.compiler_for_name = dict((name, Compiler.fromConfig(context, config))
                                  for name, config in _CONFIG_FOR_COMPILER.items())

  def _tempname(self):
    # don't assume the user's cwd is buildroot
    buildroot = get_buildroot()
    fallback = os.path.join(buildroot, '.pants.d')
    pants_workdir = self.context.config.getdefault('pants_workdir', default=fallback)
    tmp_dir = os.path.join(pants_workdir, 'tmp')
    safe_mkdir(tmp_dir)
    fd, path = tempfile.mkstemp(dir=tmp_dir, prefix='')
    os.close(fd)
    return path

  def execute(self, targets):
    self._validate(targets)

    gentargets_by_dependee = self.context.dependants(
      on_predicate=self.is_gentarget,
      from_predicate=lambda t: not self.is_gentarget(t))

    dependees_by_gentarget = defaultdict(set)
    for dependee, tgts in gentargets_by_dependee.items():
      for gentarget in tgts:
        dependees_by_gentarget[gentarget].add(dependee)

    partial_cmds = defaultdict(set)
    gentargets = filter(self.is_gentarget, targets)

    for target in gentargets:
      partial_cmd = self.PartialCmd(
        compiler=self.compiler_for_name[target.compiler],
        language=target.language,
        rpc_style=target.rpc_style,
        namespace_map=tuple(target.namespace_map.items()) if target.namespace_map else ())
      partial_cmds[partial_cmd].add(target)

    # TODO(John Sirois): This invalidates scrooge and scrooge-legacy targets even if only one
    # profile has changed as a side-effect of invoking self.profile_classpath(...).  Split up
    # ScroogeGen into 2 concrete Tasks with a thin subclassing so that each compiler gets its own
    # cache.
    for partial_cmd in partial_cmds:
      self.profile_classpath(partial_cmd.compiler.profile)

    for partial_cmd, tgts in partial_cmds.items():
      gen_files_for_source = self.gen(partial_cmd, tgts)

      outdir = partial_cmd.outdir
      langtarget_by_gentarget = {}
      for target in tgts:
        dependees = dependees_by_gentarget.get(target, [])
        langtarget_by_gentarget[target] = self.createtarget(target, dependees, outdir,
                                                            gen_files_for_source)

      genmap = self.context.products.get(partial_cmd.language)
      for gentarget, langtarget in langtarget_by_gentarget.items():
        genmap.add(gentarget, get_buildroot(), [langtarget])
        for dep in gentarget.internal_dependencies:
          if self.is_gentarget(dep):
            langtarget.update_dependencies([langtarget_by_gentarget[dep]])

  def gen(self, partial_cmd, targets):
    with self.changed(targets, invalidate_dependants=True) as changed_targets:
      compiler = partial_cmd.compiler
      import_paths, changed_srcs = compiler.calc_srcs(changed_targets, self.is_gentarget)
      outdir = partial_cmd.outdir
      if changed_srcs:
        args = []

        for import_path in import_paths:
          args.extend(['--import-path', import_path])

        args.extend(['--language', partial_cmd.language])

        for lhs, rhs in partial_cmd.namespace_map:
          args.extend(['--namespace-map', '%s=%s' % (lhs, rhs)])

        if partial_cmd.rpc_style == 'ostrich':
          args.append('--finagle')
          args.append('--ostrich')
        elif partial_cmd.rpc_style == 'finagle':
          args.append('--finagle')

        args.extend(['--dest', outdir])
        safe_mkdir(outdir)

        if not compiler.strict:
          args.append('--disable-strict')

        if compiler.verbose:
          args.append('--verbose')

        gen_file_map_path = os.path.relpath(self._tempname())
        args.extend(['--gen-file-map', gen_file_map_path])

        args.extend(changed_srcs)

        returncode = self.runjava(self.profile_classpath(compiler.profile),
                                  compiler.main,
                                  args=args,
                                  jvm_args=compiler.jvm_args)
        try:
          if 0 == returncode:
            gen_files_for_source = self.parse_gen_file_map(gen_file_map_path, outdir)
        finally:
          os.remove(gen_file_map_path)

        if 0 != returncode:
          raise TaskError('java %s ... exited non-zero (%i)' % (compiler.main, returncode))
        self.write_gen_file_map(gen_files_for_source, changed_targets, outdir)

    return self.gen_file_map(targets, outdir)

  def createtarget(self, gentarget, dependees, outdir, gen_files_for_source):
    assert self.is_gentarget(gentarget)

    def create_target(files, deps, target_type):
      return self.context.add_new_target(outdir,
                                         target_type,
                                         name=gentarget.id,
                                         provides=gentarget.provides,
                                         sources=files,
                                         dependencies=deps)

    def create_geninfo(key):
      compiler = self.compiler_for_name[gentarget.compiler]
      gen_info = self.context.config.getdict(compiler.config_section, key,
                                             default={'gen': key,
                                                      'deps': {'service': [], 'structs': []}})
      gen = gen_info['gen']
      deps = dict()
      for category, depspecs in gen_info['deps'].items():
        dependencies = OrderedSet()
        deps[category] = dependencies
        for depspec in depspecs:
          dependencies.update(self.context.resolve(depspec))
      return self.GenInfo(gen, deps)

    return self._inject_target(gentarget, dependees,
                               create_geninfo(gentarget.language),
                               gen_files_for_source,
                               create_target)

  def _inject_target(self, target, dependees, geninfo, gen_files_for_source, create_target):
    files = []
    has_service = False
    for source_file in target.sources:
      source = os.path.join(target.target_base, source_file)
      services = calculate_services(source)
      genfiles = gen_files_for_source[source]
      has_service = has_service or services
      files.extend(genfiles)
    deps = OrderedSet(geninfo.deps['service' if has_service else 'structs'])
    deps.update(target.dependencies)
    target_type = _TARGET_TYPE_FOR_LANG[target.language]
    tgt = create_target(files, deps, target_type)
    tgt.derived_from = target
    tgt.is_codegen = True
    for dependee in dependees:
      if isinstance(dependee, InternalTarget):
        dependee.update_dependencies((tgt,))
      else:
        # TODO(John Sirois): rationalize targets with dependencies.
        # JarLibrary or PythonTarget dependee on the thrift target
        dependee.dependencies.add(tgt)
    return tgt

  def parse_gen_file_map(self, gen_file_map_path, outdir):
    d = defaultdict(set)
    with open(gen_file_map_path, 'r') as deps:
      for dep in deps:
        src, cls = dep.strip().split('->')
        src = os.path.relpath(src.strip())
        cls = os.path.relpath(cls.strip(), outdir)
        d[src].add(cls)
    return d

  def gen_file_map_path_for_target(self, target, outdir):
    return os.path.join(outdir, 'gen-file-map-by-target', target.id)

  def gen_file_map_for_target(self, target, outdir):
    gen_file_map = self.gen_file_map_path_for_target(target, outdir)
    return self.parse_gen_file_map(gen_file_map, outdir)

  def gen_file_map(self, targets, outdir):
    gen_file_map = defaultdict(set)
    for target in targets:
      target_gen_file_map = self.gen_file_map_for_target(target, outdir)
      gen_file_map.update(target_gen_file_map)
    return gen_file_map

  def write_gen_file_map_for_target(self, gen_file_map, target, outdir):
    def calc_srcs(target):
      _, srcs = calculate_compile_sources([target], self.is_gentarget)
      return srcs
    with safe_open(self.gen_file_map_path_for_target(target, outdir), 'w') as f:
      for src in sorted(calc_srcs(target)):
        clss = gen_file_map[src]
        for cls in sorted(clss):
          print('%s -> %s' % (src, os.path.join(outdir, cls)), file=f)

  def write_gen_file_map(self, gen_file_map, targets, outdir):
    for target in targets:
      self.write_gen_file_map_for_target(gen_file_map, target, outdir)

  def is_gentarget(self, target):
    result = (isinstance(target, JavaThriftLibrary)
              and target.compiler in self.compiler_for_name.keys())

    if result and target.language not in self.compiler_for_name[target.compiler].langs:
      raise TaskError("%s can not generate %s" % (target.compiler, target.language))
    return result

  @staticmethod
  def _validate(targets):
    ValidateCompilerConfig = namedtuple('ValidateCompilerConfig', ['language', 'rpc_style'])

    def compiler_config(tgt):
      # Note compiler is not present in this signature. At this time
      # Scrooge and the Apache thrift generators produce identical
      # java sources, and the Apache generator does not produce scala
      # sources. As there's no permutation allowing the creation of
      # incompatible sources with the same language+rpc_style we omit
      # the compiler from the signature at this time.
      return ValidateCompilerConfig(language=tgt.language, rpc_style=tgt.rpc_style)

    mismatched_compiler_configs = defaultdict(set)

    for target in filter(lambda t: isinstance(t, JavaThriftLibrary), targets):
      mycompilerconfig = compiler_config(target)
      def collect(dep):
        if mycompilerconfig != compiler_config(dep):
          mismatched_compiler_configs[target].add(dep)
      target.walk(collect, predicate=lambda t: isinstance(t, JavaThriftLibrary))

    if mismatched_compiler_configs:
      msg = ['Thrift dependency trees must be generated with a uniform compiler configuration.\n\n']
      for tgt in sorted(mismatched_compiler_configs.keys()):
        msg.append('%s - %s\n' % (tgt, compiler_config(tgt)))
        for dep in mismatched_compiler_configs[tgt]:
          msg.append('    %s - %s\n' % (dep, compiler_config(dep)))
      raise TaskError(''.join(msg))


NAMESPACE_PARSER = re.compile(r'^\s*namespace\s+([^\s]+)\s+([^\s]+)\s*$')
TYPE_PARSER = re.compile(r'^\s*(const|enum|exception|service|struct|union)\s+([^\s{]+).*')


# TODO(John Sirois): consolidate thrift parsing to 1 pass instead of 2
def calculate_services(source):
  """Calculates the services generated for the given thrift IDL source.
  Returns an interable of services
  """

  with open(source, 'r') as thrift:
    namespaces = dict()
    types = defaultdict(set)
    for line in thrift:
      match = NAMESPACE_PARSER.match(line)
      if match:
        lang = match.group(1)
        namespace = match.group(2)
        namespaces[lang] = namespace
      else:
        match = TYPE_PARSER.match(line)
        if match:
          typename = match.group(1)
          name = match.group(2)
          types[typename].add(name)

    return types['service']



from collections import defaultdict

class Products(object):
  class ProductMapping(object):
    """
      Maps products of a given type by target.  Its assumed that all products of a given type for
      a given target are emitted to a single base directory.
    """

    def __init__(self, typename):
      self.typename = typename
      self.by_target = defaultdict(lambda: defaultdict(list))

    def empty(self):
      return len(self.by_target) == 0

    def add(self, target, basedir, product_paths=None):
      """
        Adds a mapping of products for the given target, basedir pair.

        If product_paths are specified, these will over-write any existing mapping for this target.

        If product_paths is omitted, the current mutable list of mapped products for this target
        and basedir is returned for appending.
      """
      if product_paths is not None:
        self.by_target[target][basedir].extend(product_paths)
      else:
        return self.by_target[target][basedir]

    def get(self, target):
      """
        Returns the product mapping for the given target as a tuple of (basedir, products list).
        Can return None if there is no mapping for the given target.
      """
      return self.by_target.get(target)

    def keys_for(self, basedir, file):
      """Returns the set of keys the given mapped product is registered under."""
      keys = set()
      for key, mappings in self.by_target.items():
        for mapped in mappings.get(basedir, []):
          if file == mapped:
            keys.add(key)
            break
      return keys

    def __repr__(self):
      return 'ProductMapping(%s) {\n  %s\n}' % (self.typename, '\n  '.join(
        '%s => %s\n    %s' % (str(target), basedir, outputs)
                              for target, outputs_by_basedir in self.by_target.items()
                              for basedir, outputs in outputs_by_basedir.items()))

  def __init__(self):
    self.products = {}
    self.predicates_for_type = defaultdict(list)

  def require(self, typename, predicate=None):
    """
      Registers a requirement that products of the given type by mapped.  If a target predicate is
      supplied, only targets matching the predicate are mapped.
    """
    if predicate:
      self.predicates_for_type[typename].append(predicate)
    return self.products.setdefault(typename, Products.ProductMapping(typename))

  def isrequired(self, typename):
    """
      Returns a predicate that selects targets required for the given type if mappings are
      required.  Otherwise returns None.
    """
    if typename not in self.products:
      return None
    def combine(first, second):
      return lambda target: first(target) or second(target)
    return reduce(combine, self.predicates_for_type[typename], lambda target: False)

  def get(self, typename):
    """Returns a ProductMapping for the given type name."""
    return self.require(typename)
